import asyncio
import logging
import math
import random
import time
from typing import TYPE_CHECKING

from nebula.addons.functions import print_msg_box

if TYPE_CHECKING:
    from nebula.core.network.communications import CommunicationsManager


class Mobility:
    def __init__(self, config, cm: "CommunicationsManager"):
        logging.info("Starting mobility module...")
        self.config = config
        self.cm = cm
        self.grace_time = self.config.participant["mobility_args"]["grace_time_mobility"]
        self.period = self.config.participant["mobility_args"]["change_geo_interval"]
        self.mobility = self.config.participant["mobility_args"]["mobility"]
        self.mobility_type = self.config.participant["mobility_args"]["mobility_type"]
        self.radius_federation = float(self.config.participant["mobility_args"]["radius_federation"])
        self.scheme_mobility = self.config.participant["mobility_args"]["scheme_mobility"]
        self.round_frequency = int(self.config.participant["mobility_args"]["round_frequency"])
        # Protocol to change connections based on distance
        self.max_distance_with_direct_connections = 300  # meters
        self.max_movement_random_strategy = 100  # meters
        self.max_movement_nearest_strategy = 100  # meters
        self.max_initiate_approximation = self.max_distance_with_direct_connections * 1.2
        # Network conditions based on distance
        self.network_conditions = {
            100: {"bandwidth": "5Gbps", "delay": "5ms"},
            200: {"bandwidth": "2Gbps", "delay": "50ms"},
            300: {"bandwidth": "100Mbps", "delay": "200ms"},
            float("inf"): {"bandwidth": "10Mbps", "delay": "1000ms"},
        }
        # Current network conditions of each connection {addr: {bandwidth: "5Gbps", delay: "0ms"}}
        self.current_network_conditions = {}
        # Logging box with mobility information
        mobility_msg = f"Mobility: {self.mobility}\nMobility type: {self.mobility_type}\nRadius federation: {self.radius_federation}\nScheme mobility: {self.scheme_mobility}\nEach {self.round_frequency} rounds"
        print_msg_box(msg=mobility_msg, indent=2, title="Mobility information")

    @property
    def round(self):
        return self.cm.get_round()

    async def start(self):
        asyncio.create_task(self.run_mobility())

    async def run_mobility(self):
        if not self.mobility:
            return
        await asyncio.sleep(self.grace_time)
        while True:
            await self.change_geo_location()
            await self.change_connections_based_on_distance()
            await asyncio.sleep(self.period)

    async def change_geo_location_random_strategy(self, latitude, longitude):
        logging.info("📍  Changing geo location randomly")
        # radius_in_degrees = self.radius_federation / 111000
        max_radius_in_degrees = self.max_movement_random_strategy / 111000
        radius = random.uniform(0, max_radius_in_degrees)
        angle = random.uniform(0, 2 * math.pi)
        latitude += radius * math.cos(angle)
        longitude += radius * math.sin(angle)
        await self.set_geo_location(latitude, longitude)

    async def change_geo_location_nearest_neighbor_strategy(
        self, distance, latitude, longitude, neighbor_latitude, neighbor_longitude
    ):
        logging.info("📍  Changing geo location towards the nearest neighbor")
        scale_factor = min(1, self.max_movement_nearest_strategy / distance)
        # Calcular el ángulo hacia el vecino
        angle = math.atan2(neighbor_longitude - longitude, neighbor_latitude - latitude)
        # Conversión de movimiento máximo a grados
        max_lat_change = self.max_movement_nearest_strategy / 111000  # Cambio en grados para latitud
        max_lon_change = self.max_movement_nearest_strategy / (
            111000 * math.cos(math.radians(latitude))
        )  # Cambio en grados para longitud
        # Aplicar escala y dirección
        delta_lat = max_lat_change * math.cos(angle) * scale_factor
        delta_lon = max_lon_change * math.sin(angle) * scale_factor
        # Actualizar latitud y longitud
        new_latitude = latitude + delta_lat
        new_longitude = longitude + delta_lon
        await self.set_geo_location(new_latitude, new_longitude)

    async def set_geo_location(self, latitude, longitude):
        if latitude < -90 or latitude > 90 or longitude < -180 or longitude > 180:
            # If the new location is out of bounds, we keep the old location
            latitude = self.config.participant["mobility_args"]["latitude"]
            longitude = self.config.participant["mobility_args"]["longitude"]

        self.config.participant["mobility_args"]["latitude"] = latitude
        self.config.participant["mobility_args"]["longitude"] = longitude
        logging.info(f"📍  New geo location: {latitude}, {longitude}")

    async def change_geo_location(self):
        if self.mobility and (self.mobility_type == "topology" or self.mobility_type == "both"):
            random.seed(time.time() + self.config.participant["device_args"]["idx"])
            latitude = float(self.config.participant["mobility_args"]["latitude"])
            longitude = float(self.config.participant["mobility_args"]["longitude"])

            direct_connections = await self.cm.get_direct_connections()
            undirect_connection = await self.cm.get_undirect_connections()
            if len(undirect_connection) > len(direct_connections):
                logging.info("📍  Undirect Connections is higher than Direct Connections")
                # Get neighbor closer to me
                selected_neighbor = await self.cm.get_nearest_connections(top=1)
                logging.info(f"📍  Selected neighbor: {selected_neighbor}")
                try:
                    (
                        neighbor_latitude,
                        neighbor_longitude,
                    ) = selected_neighbor.get_geolocation()
                    distance = selected_neighbor.get_neighbor_distance()
                    if distance > self.max_initiate_approximation:
                        # If the distance is too big, we move towards the neighbor
                        await self.change_geo_location_nearest_neighbor_strategy(
                            distance,
                            latitude,
                            longitude,
                            neighbor_latitude,
                            neighbor_longitude,
                        )
                    else:
                        await self.change_geo_location_random_strategy(latitude, longitude)
                except Exception as e:
                    logging.info(f"📍  Neighbor location/distance not found for {selected_neighbor.get_addr()}: {e}")
                    await self.change_geo_location_random_strategy(latitude, longitude)
            else:
                await self.change_geo_location_random_strategy(latitude, longitude)
        else:
            logging.error(f"📍  Mobility type {self.mobility_type} not implemented")
            return

    async def change_connections_based_on_distance(self):
        if self.mobility and (self.mobility_type == "topology" or self.mobility_type == "both"):
            try:
                # logging.info(f"📍  Checking connections based on distance")
                connections_topology = await self.cm.get_addrs_current_connections()
                # logging.info(f"📍  Connections of the topology: {connections_topology}")
                if len(connections_topology) < 1:
                    # logging.error(f"📍  Not enough connections for mobility")
                    return
                # Nodes that are too far away should be marked as undirected connections, and closer nodes should be marked as directed connections.
                for addr in connections_topology:
                    distance = self.cm.connections[addr].get_neighbor_distance()
                    if distance is None:
                        # If the distance is not found, we skip the node
                        continue
                    # logging.info(f"📍  Distance to node {addr}: {distance}")
                    if (
                        not self.cm.connections[addr].get_direct()
                        and distance < self.max_distance_with_direct_connections
                    ):
                        logging.info(f"📍  Node {addr} is close enough [{distance}], adding to direct connections")
                        self.cm.connections[addr].set_direct(True)
                    else:
                        # 10% margin to avoid oscillations
                        if (
                            self.cm.connections[addr].get_direct()
                            and distance > self.max_distance_with_direct_connections * 1.1
                        ):
                            logging.info(
                                f"📍  Node {addr} is too far away [{distance}], removing from direct connections"
                            )
                            self.cm.connections[addr].set_direct(False)
                    # Adapt network conditions of the connection based on distance
                    for threshold in sorted(self.network_conditions.keys()):
                        if distance < threshold:
                            conditions = self.network_conditions[threshold]
                            break
                    # Only update the network conditions if they have changed
                    if (
                        addr not in self.current_network_conditions
                        or self.current_network_conditions[addr] != conditions
                    ):
                        # eth1 is the interface of the container that connects to the node network - eth0 is the interface of the container that connects to the frontend/backend
                        self.cm._set_network_conditions(
                            interface="eth1",
                            network=addr.split(":")[0],
                            bandwidth=conditions["bandwidth"],
                            delay=conditions["delay"],
                            delay_distro="10ms",
                            delay_distribution="normal",
                            loss="0%",
                            duplicate="0%",
                            corrupt="0%",
                            reordering="0%",
                        )
                        self.current_network_conditions[addr] = conditions
            except KeyError as e:
                # Except when self.cm.connections[addr] is not found (disconnected during the process)
                logging.exception(f"📍  Connection {addr} not found: {e}")
                return
            except Exception as e:
                logging.exception(f"📍  Error changing connections based on distance: {e}")
                return

    async def change_connections(self):
        if (
            self.mobility
            and (self.mobility_type == "topology" or self.mobility_type == "both")
            and self.round % self.round_frequency == 0
        ):
            logging.info("📍  Changing connections")
            current_connections = await self.cm.get_addrs_current_connections(only_direct=True)
            potential_connections = await self.cm.get_addrs_current_connections(only_undirected=True)
            logging.info(
                f"📍  Current connections: {current_connections} | Potential future connections: {potential_connections}"
            )
            if len(current_connections) < 1 or len(potential_connections) < 1:
                logging.error("📍  Not enough connections for mobility")
                return

            if self.scheme_mobility == "random":
                random_neighbor = random.choice(current_connections)
                random_potential_neighbor = random.choice(potential_connections)
                logging.info(f"📍  Selected node(s) to disconnect: {random_neighbor}")
                logging.info(f"📍  Selected node(s) to connect: {random_potential_neighbor}")
                await self.cm.disconnect(random_neighbor, mutual_disconnection=True)
                await self.cm.connect(random_potential_neighbor, direct=True)
                logging.info(f"📍  New connections: {self.get_current_connections(only_direct=True)}")
                logging.info(f"📍  Neighbors in config: {self.config.participant['network_args']['neighbors']}")
            else:
                logging.error(f"📍  Mobility scheme {self.scheme_mobility} not implemented")
                return



import logging

import numpy as np
import torch

from nebula.core.aggregation.aggregator import Aggregator
from nebula.core.utils.helper import cosine_metric


class DualHistAgg(Aggregator):
    """
    Aggregator: Dual History Aggregation (DualHistAgg)
    Authors: Enrique et al.
    Year: 2024
    """

    def __init__(self, config=None, **kwargs):
        super().__init__(config, **kwargs)

    def softmax(self, x):
        # Safeguard against empty input array
        if x.size == 0:
            return np.array([])
        e_x = np.exp(x - np.max(x))
        return e_x / e_x.sum(axis=0)  # ensure division is done correctly

    def run_aggregation(self, models, reference_model=None):
        if len(models) == 0:
            logging.error("Trying to aggregate models when there are no models")
            return None, None

        models = list(models.values())
        num_models = len(models)
        logging.info(f"Number of models: {num_models}")

        if num_models == 1:
            logging.info("Only one model, returning it")
            return models[0][0], models[0][0]

        # Total Samples
        total_samples = float(sum(w for _, w in models))
        # Create a Zero Model
        accum = {
            layer: torch.zeros_like(param).float() for layer, param in models[0][0].items()
        }  # use first model for template
        accum_similarity = accum.copy()

        similarities = (
            [cosine_metric(model, reference_model) for model, _ in models] if reference_model else [1] * num_models
        )

        logging.info(f"Similarities: {similarities}")
        weights = self.softmax(np.array(similarities))
        logging.info(f"Weights: {weights}")

        # Aggregation process
        for (model, _), weight, sim_weight in zip(models, weights, similarities, strict=False):
            for layer in accum:
                accum[layer] += model[layer].float() * float(weight)
                accum_similarity[layer] += model[layer].float() * float(sim_weight)

        # Normalize aggregated models
        for layer in accum:
            accum[layer] /= total_samples
            accum_similarity[layer] /= total_samples

        return accum, accum_similarity



import logging
from abc import ABC, abstractmethod
from functools import partial

from nebula.core.pb import nebula_pb2
from nebula.core.utils.locker import Locker


class AggregatorException(Exception):
    pass


def create_aggregator(config, engine):
    from nebula.core.aggregation.blockchainReputation import BlockchainReputation
    from nebula.core.aggregation.fedavg import FedAvg
    from nebula.core.aggregation.krum import Krum
    from nebula.core.aggregation.median import Median
    from nebula.core.aggregation.trimmedmean import TrimmedMean

    ALGORITHM_MAP = {
        "FedAvg": FedAvg,
        "Krum": Krum,
        "Median": Median,
        "TrimmedMean": TrimmedMean,
        "BlockchainReputation": BlockchainReputation,
    }
    algorithm = config.participant["aggregator_args"]["algorithm"]
    aggregator = ALGORITHM_MAP.get(algorithm)
    if aggregator:
        return aggregator(config=config, engine=engine)
    else:
        raise AggregatorException(f"Aggregation algorithm {algorithm} not found.")


def create_target_aggregator(config, engine):
    from nebula.core.aggregation.fedavg import FedAvg
    from nebula.core.aggregation.krum import Krum
    from nebula.core.aggregation.median import Median
    from nebula.core.aggregation.trimmedmean import TrimmedMean

    ALGORITHM_MAP = {
        "FedAvg": FedAvg,
        "Krum": Krum,
        "Median": Median,
        "TrimmedMean": TrimmedMean,
    }
    algorithm = config.participant["defense_args"]["target_aggregation"]
    aggregator = ALGORITHM_MAP.get(algorithm)
    if aggregator:
        return aggregator(config=config, engine=engine)
    else:
        raise AggregatorException(f"Aggregation algorithm {algorithm} not found.")


class Aggregator(ABC):
    def __init__(self, config=None, engine=None):
        self.config = config
        self.engine = engine
        self._addr = config.participant["network_args"]["addr"]
        logging.info(f"[{self.__class__.__name__}] Starting Aggregator")
        self._federation_nodes = set()
        self._waiting_global_update = False
        self._pending_models_to_aggregate = {}
        self._future_models_to_aggregate = {}
        self._add_model_lock = Locker(name="add_model_lock", async_lock=True)
        self._aggregation_done_lock = Locker(name="aggregation_done_lock", async_lock=True)

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.__str__()

    @property
    def cm(self):
        return self.engine.cm

    @abstractmethod
    def run_aggregation(self, models):
        if len(models) == 0:
            logging.error("Trying to aggregate models when there are no models")
            return None

    async def update_federation_nodes(self, federation_nodes):
        if not self._aggregation_done_lock.locked():
            self._federation_nodes = federation_nodes
            self._pending_models_to_aggregate.clear()
            await self._aggregation_done_lock.acquire_async(
                timeout=self.config.participant["aggregator_args"]["aggregation_timeout"]
            )
        else:
            raise Exception("It is not possible to set nodes to aggregate when the aggregation is running.")

    def set_waiting_global_update(self):
        self._waiting_global_update = True

    async def reset(self):
        await self._add_model_lock.acquire_async()
        self._federation_nodes.clear()
        self._pending_models_to_aggregate.clear()
        try:
            await self._aggregation_done_lock.release_async()
        except:
            pass
        await self._add_model_lock.release_async()

    def get_nodes_pending_models_to_aggregate(self):
        return {node for key in self._pending_models_to_aggregate.keys() for node in key.split()}

    async def _handle_global_update(self, model, source):
        logging.info(f"🔄  _handle_global_update | source={source}")
        logging.info(
            f"🔄  _handle_global_update | Received a model from {source}. Overwriting __models with the aggregated model."
        )
        self._pending_models_to_aggregate.clear()
        self._pending_models_to_aggregate = {source: (model, 1)}
        self._waiting_global_update = False
        await self._add_model_lock.release_async()
        await self._aggregation_done_lock.release_async()

    async def _add_pending_model(self, model, weight, source):
        if len(self._federation_nodes) <= len(self.get_nodes_pending_models_to_aggregate()):
            logging.info("🔄  _add_pending_model | Ignoring model...")
            await self._add_model_lock.release_async()
            return None

        if source not in self._federation_nodes:
            logging.info(f"🔄  _add_pending_model | Can't add a model from ({source}), which is not in the federation.")
            await self._add_model_lock.release_async()
            return None

        elif source not in self.get_nodes_pending_models_to_aggregate():
            logging.info(
                "🔄  _add_pending_model | Node is not in the aggregation buffer --> Include model in the aggregation buffer."
            )
            self._pending_models_to_aggregate.update({source: (model, weight)})

        logging.info(
            f"🔄  _add_pending_model | Model added in aggregation buffer ({len(self.get_nodes_pending_models_to_aggregate())!s}/{len(self._federation_nodes)!s}) | Pending nodes: {self._federation_nodes - self.get_nodes_pending_models_to_aggregate()}"
        )

        # Check if _future_models_to_aggregate has models in the current round to include in the aggregation buffer
        if self.engine.get_round() in self._future_models_to_aggregate:
            logging.info(
                f"🔄  _add_pending_model | Including next models in the aggregation buffer for round {self.engine.get_round()}"
            )
            for future_model in self._future_models_to_aggregate[self.engine.get_round()]:
                if future_model is None:
                    continue
                future_model, future_weight, future_source = future_model
                if (
                    future_source in self._federation_nodes
                    and future_source not in self.get_nodes_pending_models_to_aggregate()
                ):
                    self._pending_models_to_aggregate.update({future_source: (future_model, future_weight)})
                    logging.info(
                        f"🔄  _add_pending_model | Next model added in aggregation buffer ({len(self.get_nodes_pending_models_to_aggregate())!s}/{len(self._federation_nodes)!s}) | Pending nodes: {self._federation_nodes - self.get_nodes_pending_models_to_aggregate()}"
                    )
            del self._future_models_to_aggregate[self.engine.get_round()]

            for future_round in list(self._future_models_to_aggregate.keys()):
                if future_round < self.engine.get_round():
                    del self._future_models_to_aggregate[future_round]

        if len(self.get_nodes_pending_models_to_aggregate()) >= len(self._federation_nodes):
            logging.info("🔄  _add_pending_model | All models were added in the aggregation buffer. Run aggregation...")
            await self._aggregation_done_lock.release_async()
        await self._add_model_lock.release_async()
        return self.get_nodes_pending_models_to_aggregate()

    async def include_model_in_buffer(self, model, weight, source=None, round=None, local=False):
        await self._add_model_lock.acquire_async()
        logging.info(
            f"🔄  include_model_in_buffer | source={source} | round={round} | weight={weight} |--| __models={self._pending_models_to_aggregate.keys()} | federation_nodes={self._federation_nodes} | pending_models_to_aggregate={self.get_nodes_pending_models_to_aggregate()}"
        )
        if model is None:
            logging.info("🔄  include_model_in_buffer | Ignoring model bad formed...")
            await self._add_model_lock.release_async()
            return

        if round == -1:
            # Be sure that the model message is not from the initialization round (round = -1)
            logging.info("🔄  include_model_in_buffer | Ignoring model with round -1")
            await self._add_model_lock.release_async()
            return

        if self._waiting_global_update and not local:
            await self._handle_global_update(model, source)
            return

        await self._add_pending_model(model, weight, source)

        if len(self.get_nodes_pending_models_to_aggregate()) >= len(self._federation_nodes):
            logging.info(
                f"🔄  include_model_in_buffer | Broadcasting MODELS_INCLUDED for round {self.engine.get_round()}"
            )
            message = self.cm.mm.generate_federation_message(
                nebula_pb2.FederationMessage.Action.FEDERATION_MODELS_INCLUDED,
                [self.engine.get_round()],
            )
            await self.cm.send_message_to_neighbors(message)

        return

    async def get_aggregation(self):
        try:
            timeout = self.config.participant["aggregator_args"]["aggregation_timeout"]
            await self._aggregation_done_lock.acquire_async(timeout=timeout)
        except TimeoutError:
            logging.exception("🔄  get_aggregation | Timeout reached for aggregation")
        finally:
            await self._aggregation_done_lock.release_async()

        if self._waiting_global_update and len(self._pending_models_to_aggregate) == 1:
            logging.info(
                "🔄  get_aggregation | Received an global model. Overwriting my model with the aggregated model."
            )
            aggregated_model = next(iter(self._pending_models_to_aggregate.values()))[0]
            self._pending_models_to_aggregate.clear()
            return aggregated_model

        unique_nodes_involved = set(node for key in self._pending_models_to_aggregate for node in key.split())

        if len(unique_nodes_involved) != len(self._federation_nodes):
            missing_nodes = self._federation_nodes - unique_nodes_involved
            logging.info(f"🔄  get_aggregation | Aggregation incomplete, missing models from: {missing_nodes}")
        else:
            logging.info("🔄  get_aggregation | All models accounted for, proceeding with aggregation.")

        aggregated_result = self.run_aggregation(self._pending_models_to_aggregate)
        self._pending_models_to_aggregate.clear()
        return aggregated_result

    async def include_next_model_in_buffer(self, model, weight, source=None, round=None):
        logging.info(f"🔄  include_next_model_in_buffer | source={source} | round={round} | weight={weight}")
        if round not in self._future_models_to_aggregate:
            self._future_models_to_aggregate[round] = []
        decoded_model = self.engine.trainer.deserialize_model(model)
        self._future_models_to_aggregate[round].append((decoded_model, weight, source))

    def print_model_size(self, model):
        total_params = 0
        total_memory = 0

        for _, param in model.items():
            num_params = param.numel()
            total_params += num_params

            memory_usage = param.element_size() * num_params
            total_memory += memory_usage

        total_memory_in_mb = total_memory / (1024**2)
        logging.info(f"print_model_size | Model size: {total_memory_in_mb} MB")


def create_malicious_aggregator(aggregator, attack):
    # It creates a partial function aggregate that wraps the aggregate method of the original aggregator.
    run_aggregation = partial(aggregator.run_aggregation)  # None is the self (not used)

    # This function will replace the original aggregate method of the aggregator.
    def malicious_aggregate(self, models):
        accum = run_aggregation(models)
        logging.info(f"malicious_aggregate | original aggregation result={accum}")
        if models is not None:
            accum = attack(accum)
            logging.info(f"malicious_aggregate | attack aggregation result={accum}")
        return accum

    aggregator.run_aggregation = partial(malicious_aggregate, aggregator)
    return aggregator



import torch

from nebula.core.models.nebulamodel import NebulaModel


class EMNISTModelMLP(NebulaModel):
    def __init__(
        self,
        input_channels=1,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}

        self.example_input_array = torch.zeros(1, 1, 28, 28)
        self.learning_rate = learning_rate
        self.criterion = torch.nn.CrossEntropyLoss()

        self.l1 = torch.nn.Linear(28 * 28, 256)
        self.l2 = torch.nn.Linear(256, 128)
        self.l3 = torch.nn.Linear(128, num_classes)

    def forward(self, x):
        batch_size, channels, width, height = x.size()
        x = x.view(batch_size, -1)
        x = self.l1(x)
        x = torch.relu(x)
        x = self.l2(x)
        x = torch.relu(x)
        x = self.l3(x)
        return x

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



import argparse
import asyncio
import datetime
import io
import json
import logging
import os
import signal
import sys
import zipfile
from urllib.parse import urlencode

import requests

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", ".."))

logging.basicConfig(level=logging.INFO)


from ansi2html import Ansi2HTMLConverter
from fastapi import (
    BackgroundTasks,
    Depends,
    FastAPI,
    Form,
    HTTPException,
    Request,
    Response,
    WebSocket,
    WebSocketDisconnect,
    status,
)
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    StreamingResponse,
)
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.middleware.sessions import SessionMiddleware

from nebula.frontend.database import (
    add_user,
    check_scenario_with_role,
    delete_user_from_db,
    get_all_scenarios_and_check_completed,
    get_notes,
    get_run_hashes_scenario,
    get_running_scenario,
    get_scenario_by_name,
    get_user_info,
    initialize_databases,
    list_nodes_by_scenario_name,
    list_users,
    remove_nodes_by_scenario_name,
    remove_note,
    remove_scenario_by_name,
    save_notes,
    scenario_set_all_status_to_finished,
    scenario_set_status_to_finished,
    scenario_update_record,
    update_node_record,
    update_user,
    verify,
    verify_hash_algorithm,
)
from nebula.frontend.utils import Utils


class Settings:
    port: int = os.environ.get("NEBULA_FRONTEND_PORT", 6060)
    production: bool = os.environ.get("NEBULA_PRODUCTION", "False") == "True"
    gpu_available: bool = os.environ.get("NEBULA_GPU_AVAILABLE", "False") == "True"
    advanced_analytics: bool = os.environ.get("NEBULA_ADVANCED_ANALYTICS", "False") == "True"
    host_platform: str = os.environ.get("NEBULA_HOST_PLATFORM", "unix")
    log_dir: str = os.environ.get("NEBULA_LOGS_DIR")
    config_dir: str = os.environ.get("NEBULA_CONFIG_DIR")
    cert_dir: str = os.environ.get("NEBULA_CERTS_DIR")
    root_host_path: str = os.environ.get("NEBULA_ROOT_HOST")
    config_frontend_dir: str = os.environ.get("FEDELLAR_CONFIG_FRONTEND_DIR", "config")
    statistics_port: int = os.environ.get("NEBULA_STATISTICS_PORT", 8080)
    secret_key: str = os.environ.get("SECRET_KEY", os.urandom(24).hex())
    PERMANENT_SESSION_LIFETIME: datetime.timedelta = datetime.timedelta(minutes=60)
    templates_dir: str = "templates"


settings = Settings()

logging.info(f"NEBULA_PRODUCTION: {settings.production}")
logging.info(f"NEBULA_ADVANCED_ANALYTICS: {settings.advanced_analytics}")

app = FastAPI()
app.add_middleware(SessionMiddleware, secret_key=settings.secret_key)
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.mount("/nebula/static", StaticFiles(directory="static"), name="static")


class ConnectionManager:
    def __init__(self):
        self.active_connections: list[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)
        message = {
            "type": "control",
            "message": f"Client #{len(self.active_connections)} connected",
        }
        await self.broadcast(json.dumps(message))

    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)

    async def send_personal_message(self, message: str, websocket: WebSocket):
        await websocket.send_text(message)

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await connection.send_text(message)


manager = ConnectionManager()


@app.websocket("/nebula/ws/{client_id}")
async def websocket_endpoint(websocket: WebSocket, client_id: int):
    await manager.connect(websocket)
    try:
        while True:
            data = await websocket.receive_text()
            message = {
                "type": "control",
                "message": f"Client #{client_id} says: {data}",
            }
            await manager.broadcast(json.dumps(message))
            # await manager.send_personal_message(f"You wrote: {data}", websocket)
    except WebSocketDisconnect:
        manager.disconnect(websocket)
        message = {"type": "control", "message": f"Client #{client_id} left the chat"}
        await manager.broadcast(json.dumps(message))


templates = Jinja2Templates(directory=settings.templates_dir)


def datetimeformat(value, format="%B %d, %Y %H:%M"):
    return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S").strftime(format)


def add_global_context(request: Request):
    return {
        "is_production": settings.production,
    }


templates.env.filters["datetimeformat"] = datetimeformat
templates.env.globals.update(add_global_context=add_global_context)


def get_session(request: Request) -> dict:
    return request.session


def set_default_user():
    username = os.environ.get("NEBULA_DEFAULT_USER", "admin")
    password = os.environ.get("NEBULA_DEFAULT_PASSWORD", "admin")
    if not list_users():
        add_user(username, password, "admin")
    if not verify_hash_algorithm(username):
        update_user(username, password, "admin")


@app.on_event("startup")
async def startup_event():
    await initialize_databases()
    set_default_user()


nodes_registration = {}

scenarios_list_length = 0

scenarios_finished = 0


# Detect CTRL+C from parent process
def signal_handler(signal, frame):
    logging.info("You pressed Ctrl+C [frontend]!")
    scenario_set_all_status_to_finished()
    sys.exit(0)


signal.signal(signal.SIGINT, signal_handler)


@app.exception_handler(StarletteHTTPException)
async def custom_http_exception_handler(request: Request, exc: StarletteHTTPException):
    context = {"request": request, "session": request.session}
    if exc.status_code == status.HTTP_401_UNAUTHORIZED:
        return templates.TemplateResponse("401.html", context, status_code=exc.status_code)
    elif exc.status_code == status.HTTP_403_FORBIDDEN:
        return templates.TemplateResponse("403.html", context, status_code=exc.status_code)
    elif exc.status_code == status.HTTP_404_NOT_FOUND:
        return templates.TemplateResponse("404.html", context, status_code=exc.status_code)
    elif exc.status_code == status.HTTP_405_METHOD_NOT_ALLOWED:
        return templates.TemplateResponse("405.html", context, status_code=exc.status_code)
    elif exc.status_code == status.HTTP_413_REQUEST_ENTITY_TOO_LARGE:
        return templates.TemplateResponse("413.html", context, status_code=exc.status_code)
    return await request.app.default_exception_handler(request, exc)


@app.get("/", response_class=HTMLResponse)
async def index():
    return RedirectResponse(url="/nebula")


@app.get("/nebula", response_class=HTMLResponse)
@app.get("/nebula/", response_class=HTMLResponse)
async def nebula_home(request: Request):
    alerts = []
    return templates.TemplateResponse("index.html", {"request": request, "alerts": alerts})


@app.get("/nebula/dashboard/{scenario_name}/private", response_class=HTMLResponse)
async def nebula_dashboard_private(request: Request, scenario_name: str, session: dict = Depends(get_session)):
    if "user" in session:
        return templates.TemplateResponse("private.html", {"request": request, "scenario_name": scenario_name})
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


@app.get("/nebula/admin", response_class=HTMLResponse)
async def nebula_admin(request: Request, session: dict = Depends(get_session)):
    if session.get("role") == "admin":
        user_list = list_users(all_info=True)
        user_table = zip(
            range(1, len(user_list) + 1),
            [user[0] for user in user_list],
            [user[2] for user in user_list],
            strict=False,
        )
        return templates.TemplateResponse("admin.html", {"request": request, "users": user_table})
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


@app.post("/nebula/dashboard/{scenario_name}/save_note")
async def save_note_for_scenario(scenario_name: str, request: Request, session: dict = Depends(get_session)):
    if "user" in session:
        data = await request.json()
        notes = data["notes"]
        try:
            save_notes(scenario_name, notes)
            return JSONResponse({"status": "success"})
        except Exception as e:
            logging.exception(e)
            return JSONResponse(
                {"status": "error", "message": "Could not save the notes"},
                status_code=500,
            )
    else:
        return JSONResponse({"status": "error", "message": "User not logged in"}, status_code=401)


@app.get("/nebula/dashboard/{scenario_name}/notes")
async def get_notes_for_scenario(scenario_name: str):
    notes_record = get_notes(scenario_name)
    if notes_record:
        notes_data = dict(zip(notes_record.keys(), notes_record, strict=False))
        return JSONResponse({"status": "success", "notes": notes_data["scenario_notes"]})
    else:
        return JSONResponse({"status": "error", "message": "Notes not found for the specified scenario"})


@app.post("/nebula/login")
async def nebula_login(
    request: Request,
    session: dict = Depends(get_session),
    user: str = Form(...),
    password: str = Form(...),
):
    user_submitted = user.upper()
    if (user_submitted in list_users()) and verify(user_submitted, password):
        user_info = get_user_info(user_submitted)
        session["user"] = user_submitted
        session["role"] = user_info[2]
        return JSONResponse({"message": "Login successful"}, status_code=200)
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


@app.get("/nebula/logout")
async def nebula_logout(request: Request, session: dict = Depends(get_session)):
    session.pop("user", None)
    return RedirectResponse(url="/nebula")


@app.get("/nebula/user/delete/{user}/")
async def nebula_delete_user(user: str, request: Request, session: dict = Depends(get_session)):
    if session.get("role") == "admin":
        if user == "ADMIN":  # ADMIN account can't be deleted.
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)
        if user == session["user"]:  # Current user can't delete himself.
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)

        delete_user_from_db(user)
        return RedirectResponse(url="/nebula/admin")
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


@app.post("/nebula/user/add")
async def nebula_add_user(
    request: Request,
    session: dict = Depends(get_session),
    user: str = Form(...),
    password: str = Form(...),
    role: str = Form(...),
):
    if session.get("role") == "admin":  # only Admin should be able to add user.
        user_list = list_users(all_info=True)
        if user.upper() in user_list or " " in user or "'" in user or '"' in user:
            return RedirectResponse(url="/nebula/admin")
        else:
            add_user(user, password, role)
            return RedirectResponse(url="/nebula/admin")
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


@app.post("/nebula/user/update")
async def nebula_update_user(
    request: Request,
    session: dict = Depends(get_session),
    user: str = Form(...),
    password: str = Form(...),
    role: str = Form(...),
):
    if session.get("role") == "admin":
        user_list = list_users()
        if user not in user_list:
            return RedirectResponse(url="/nebula/admin")
        else:
            update_user(user, password, role)
            return RedirectResponse(url="/nebula/admin")
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


@app.get("/nebula/api/dashboard/runningscenario", response_class=JSONResponse)
async def nebula_dashboard_runningscenario():
    scenario_running = get_running_scenario()
    if scenario_running:
        scenario_running_as_dict = dict(scenario_running)
        scenario_running_as_dict["scenario_status"] = "running"
        return JSONResponse(scenario_running_as_dict)
    else:
        return JSONResponse({"scenario_status": "not running"})


@app.get("/nebula/api/dashboard", response_class=JSONResponse)
@app.get("/nebula/dashboard", response_class=HTMLResponse)
async def nebula_dashboard(request: Request, session: dict = Depends(get_session)):
    if "user" in session:
        scenarios = get_all_scenarios_and_check_completed()  # Get all scenarios after checking if they are completed
        scenario_running = get_running_scenario()
    else:
        scenarios = None
        scenario_running = None

    bool_completed = False
    if scenario_running:
        bool_completed = scenario_running[5] == "completed"

    if scenarios:
        if request.url.path == "/nebula/dashboard":
            return templates.TemplateResponse(
                "dashboard.html",
                {
                    "request": request,
                    "scenarios": scenarios,
                    "scenarios_list_length": scenarios_list_length,
                    "scenarios_finished": scenarios_finished,
                    "scenario_running": scenario_running,
                    "scenario_completed": bool_completed,
                    "user_logged_in": session.get("user"),
                },
            )
        elif request.url.path == "/nebula/api/dashboard":
            scenarios_as_dict = [dict(row) for row in scenarios]
            return JSONResponse(scenarios_as_dict)
        else:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
    else:
        if request.url.path == "/nebula/dashboard":
            return templates.TemplateResponse(
                "dashboard.html",
                {
                    "request": request,
                    "user_logged_in": session.get("user"),
                },
            )
        elif request.url.path == "/nebula/api/dashboard":
            return JSONResponse({"scenarios_status": "not found in database"})
        else:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


@app.get("/nebula/api/dashboard/{scenario_name}/monitor", response_class=JSONResponse)
@app.get("/nebula/dashboard/{scenario_name}/monitor", response_class=HTMLResponse)
async def nebula_dashboard_monitor(scenario_name: str, request: Request, session: dict = Depends(get_session)):
    scenario = get_scenario_by_name(scenario_name)
    if scenario:
        nodes_list = list_nodes_by_scenario_name(scenario_name)
        if nodes_list:
            nodes_config = []
            nodes_status = []
            for node in nodes_list:
                nodes_config.append((node[2], node[3], node[4]))  # IP, Port, Role
                if datetime.datetime.now() - datetime.datetime.strptime(
                    node[8], "%Y-%m-%d %H:%M:%S.%f"
                ) > datetime.timedelta(seconds=25):
                    nodes_status.append(False)
                else:
                    nodes_status.append(True)
            nodes_table = zip(
                [x[0] for x in nodes_list],  # UID
                [x[1] for x in nodes_list],  # IDX
                [x[2] for x in nodes_list],  # IP
                [x[3] for x in nodes_list],  # Port
                [x[4] for x in nodes_list],  # Role
                [x[5] for x in nodes_list],  # Neighbors
                [x[6] for x in nodes_list],  # Latitude
                [x[7] for x in nodes_list],  # Longitude
                [x[8] for x in nodes_list],  # Timestamp
                [x[9] for x in nodes_list],  # Federation
                [x[10] for x in nodes_list],  # Round
                [x[11] for x in nodes_list],  # Scenario name
                [x[12] for x in nodes_list],  # Run hash
                nodes_status,
                strict=False,  # Status
            )

            topology_path = Utils.check_path(settings.config_dir, os.path.join(scenario_name, "topology.png"))
            if os.path.exists(topology_path):
                latest_participant_file_mtime = max([
                    os.path.getmtime(
                        os.path.join(
                            settings.config_dir,
                            scenario_name,
                            f"participant_{node[1]}.json",
                        )
                    )
                    for node in nodes_list
                ])
                if os.path.getmtime(topology_path) < latest_participant_file_mtime:
                    update_topology(scenario[0], nodes_list, nodes_config)
            else:
                update_topology(scenario[0], nodes_list, nodes_config)

            if request.url.path == f"/nebula/dashboard/{scenario_name}/monitor":
                return templates.TemplateResponse(
                    "monitor.html",
                    {
                        "request": request,
                        "scenario_name": scenario_name,
                        "scenario": scenario,
                        "nodes": nodes_table,
                        "user_logged_in": session.get("user"),
                    },
                )
            elif request.url.path == f"/nebula/api/dashboard/{scenario_name}/monitor":
                return JSONResponse({
                    "scenario_status": scenario[5],
                    "nodes_table": list(nodes_table),
                    "scenario_name": scenario[0],
                    "scenario_title": scenario[3],
                    "scenario_description": scenario[4],
                })
            else:
                raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
        else:
            if request.url.path == f"/nebula/dashboard/{scenario_name}/monitor":
                return templates.TemplateResponse(
                    "monitor.html",
                    {
                        "request": request,
                        "scenario_name": scenario_name,
                        "scenario": scenario,
                        "nodes": [],
                        "user_logged_in": session.get("user"),
                    },
                )
            elif request.url.path == f"/nebula/api/dashboard/{scenario_name}/monitor":
                return JSONResponse({
                    "scenario_status": scenario[5],
                    "nodes_table": [],
                    "scenario_name": scenario[0],
                    "scenario_title": scenario[3],
                    "scenario_description": scenario[4],
                })
            else:
                raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
    else:
        if request.url.path == f"/nebula/dashboard/{scenario_name}/monitor":
            return templates.TemplateResponse(
                "monitor.html",
                {
                    "request": request,
                    "scenario_name": scenario_name,
                    "scenario": None,
                    "nodes": [],
                    "user_logged_in": session.get("user"),
                },
            )
        elif request.url.path == f"/nebula/api/dashboard/{scenario_name}/monitor":
            return JSONResponse({"scenario_status": "not exists"})
        else:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


def update_topology(scenario_name, nodes_list, nodes_config):
    import numpy as np

    nodes = []
    for node in nodes_list:
        nodes.append(node[2] + ":" + str(node[3]))
    matrix = np.zeros((len(nodes), len(nodes)))
    for node in nodes_list:
        for neighbour in node[5].split(" "):
            if neighbour != "":
                if neighbour in nodes:
                    matrix[
                        nodes.index(node[2] + ":" + str(node[3])),
                        nodes.index(neighbour),
                    ] = 1
    from nebula.addons.topologymanager import TopologyManager

    tm = TopologyManager(n_nodes=len(nodes_list), topology=matrix, scenario_name=scenario_name)
    tm.update_nodes(nodes_config)
    tm.draw_graph(path=os.path.join(settings.config_dir, scenario_name, "topology.png"))


@app.post("/nebula/dashboard/{scenario_name}/node/update")
async def nebula_update_node(scenario_name: str, request: Request, session: dict = Depends(get_session)):
    if request.method == "POST":
        if request.headers.get("content-type") == "application/json":
            config = await request.json()
            timestamp = datetime.datetime.now()
            # Update the node in database
            await update_node_record(
                str(config["device_args"]["uid"]),
                str(config["device_args"]["idx"]),
                str(config["network_args"]["ip"]),
                str(config["network_args"]["port"]),
                str(config["device_args"]["role"]),
                str(config["network_args"]["neighbors"]),
                str(config["mobility_args"]["latitude"]),
                str(config["mobility_args"]["longitude"]),
                str(timestamp),
                str(config["scenario_args"]["federation"]),
                str(config["federation_args"]["round"]),
                str(config["scenario_args"]["name"]),
                str(config["tracking_args"]["run_hash"]),
            )

            neighbors_distance = config["mobility_args"]["neighbors_distance"]

            node_update = {
                "type": "node_update",
                "scenario_name": scenario_name,
                "uid": config["device_args"]["uid"],
                "idx": config["device_args"]["idx"],
                "ip": config["network_args"]["ip"],
                "port": str(config["network_args"]["port"]),
                "role": config["device_args"]["role"],
                "neighbors": config["network_args"]["neighbors"],
                "latitude": config["mobility_args"]["latitude"],
                "longitude": config["mobility_args"]["longitude"],
                "timestamp": str(timestamp),
                "federation": config["scenario_args"]["federation"],
                "round": config["federation_args"]["round"],
                "name": config["scenario_args"]["name"],
                "status": True,
                "neighbors_distance": neighbors_distance,
            }

            try:
                await manager.broadcast(json.dumps(node_update))
            except Exception as e:
                logging.exception(f"Error sending node_update to socketio: {e}")
                pass

            return JSONResponse({"message": "Node updated", "status": "success"}, status_code=200)
        else:
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST)


@app.post("/nebula/dashboard/{scenario_name}/node/register")
async def nebula_register_node(scenario_name: str, request: Request):
    if request.headers.get("content-type") == "application/json":
        data = await request.json()
        node = data["node"]
        logging.info(f"Registering node {node} for scenario {scenario_name}")
        async with nodes_registration[scenario_name]["condition"]:
            nodes_registration[scenario_name]["nodes"].add(node)
            logging.info(f"Node {node} registered")
            if len(nodes_registration[scenario_name]["nodes"]) == nodes_registration[scenario_name]["n_nodes"]:
                nodes_registration[scenario_name]["condition"].notify_all()
                logging.info("All nodes registered")

        return JSONResponse({"message": "Node registered", "status": "success"}, status_code=200)
    else:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST)


@app.get("/nebula/dashboard/scenarios/node/list")
async def nebula_list_all_scenarios(session: dict = Depends(get_session)):
    if "user" not in session or session["role"] not in ["admin", "user"]:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized")

    scenarios = {}
    for scenario_name, scenario_info in nodes_registration.items():
        scenarios[scenario_name] = list(scenario_info["nodes"])

    if not scenarios:
        return JSONResponse({"message": "No scenarios found", "status": "error"}, status_code=404)

    return JSONResponse({"scenarios": scenarios, "status": "success"}, status_code=200)


@app.get("/nebula/dashboard/scenarios/node/erase")
async def nebula_erase_all_nodes(session: dict = Depends(get_session)):
    if "user" not in session or session["role"] not in ["admin", "user"]:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized")

    nodes_registration.clear()
    return JSONResponse({"message": "All nodes erased", "status": "success"}, status_code=200)


@app.get("/nebula/dashboard/{scenario_name}/node/wait")
async def nebula_wait_nodes(scenario_name: str):
    if scenario_name not in nodes_registration:
        return JSONResponse({"message": "Scenario not found", "status": "error"}, status_code=404)

    async with nodes_registration[scenario_name]["condition"]:
        while len(nodes_registration[scenario_name]["nodes"]) < nodes_registration[scenario_name]["n_nodes"]:
            await nodes_registration[scenario_name]["condition"].wait()
        return JSONResponse({"message": "All nodes registered", "status": "success"}, status_code=200)


@app.get("/nebula/dashboard/{scenario_name}/node/{id}/infolog")
async def nebula_monitor_log(scenario_name: str, id: str):
    logs = Utils.check_path(settings.log_dir, os.path.join(scenario_name, f"participant_{id}.log"))
    if os.path.exists(logs):
        return FileResponse(logs, media_type="text/plain", filename=f"participant_{id}.log")
    else:
        raise HTTPException(status_code=404, detail="Log file not found")


@app.get(
    "/nebula/dashboard/{scenario_name}/node/{id}/infolog/{number}",
    response_class=PlainTextResponse,
)
async def nebula_monitor_log_x(scenario_name: str, id: str, number: int):
    logs = Utils.check_path(settings.log_dir, os.path.join(scenario_name, f"participant_{id}.log"))
    if os.path.exists(logs):
        with open(logs) as f:
            lines = f.readlines()[-number:]
            lines = "".join(lines)
            converter = Ansi2HTMLConverter()
            html_text = converter.convert(lines, full=False)
            return Response(content=html_text, media_type="text/plain")
    else:
        return Response(content="No logs available", media_type="text/plain")


@app.get("/nebula/dashboard/{scenario_name}/node/{id}/debuglog")
async def nebula_monitor_log_debug(scenario_name: str, id: str):
    logs = Utils.check_path(settings.log_dir, os.path.join(scenario_name, f"participant_{id}_debug.log"))
    if os.path.exists(logs):
        return FileResponse(logs, media_type="text/plain", filename=f"participant_{id}_debug.log")
    else:
        raise HTTPException(status_code=404, detail="Log file not found")


@app.get("/nebula/dashboard/{scenario_name}/node/{id}/errorlog")
async def nebula_monitor_log_error(scenario_name: str, id: str):
    logs = Utils.check_path(settings.log_dir, os.path.join(scenario_name, f"participant_{id}_error.log"))
    if os.path.exists(logs):
        return FileResponse(logs, media_type="text/plain", filename=f"participant_{id}_error.log")
    else:
        raise HTTPException(status_code=404, detail="Log file not found")


@app.get("/nebula/dashboard/{scenario_name}/topology/image/")
async def nebula_monitor_image(scenario_name: str):
    topology_image = Utils.check_path(settings.log_dir, os.path.join(scenario_name, "topology.png"))
    if os.path.exists(topology_image):
        return FileResponse(topology_image, media_type="image/png")
    else:
        raise HTTPException(status_code=404, detail="Topology image not found")


def stop_scenario(scenario_name):
    from nebula.scenarios import ScenarioManagement

    ScenarioManagement.stop_participants()
    ScenarioManagement.stop_blockchain()
    scenario_set_status_to_finished(scenario_name)


def stop_all_scenarios():
    from nebula.scenarios import ScenarioManagement

    ScenarioManagement.stop_participants()
    ScenarioManagement.stop_blockchain()
    scenario_set_all_status_to_finished()


@app.get("/nebula/dashboard/{scenario_name}/stop/{stop_all}")
async def nebula_stop_scenario(
    scenario_name: str,
    stop_all: bool,
    request: Request,
    session: dict = Depends(get_session),
):
    if "user" in session:
        if session["role"] == "demo":
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
        elif session["role"] == "user":
            if not check_scenario_with_role(session["role"], scenario_name):
                raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
        if stop_all:
            stop_all_scenarios_event.set()
            global scenarios_list_length
            global scenarios_finished
            scenarios_list_length = 0
            scenarios_finished = 0
            stop_scenario(scenario_name)
        else:
            finish_scenario_event.set()
            stop_scenario(scenario_name)
        return RedirectResponse(url="/nebula/dashboard")
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


def remove_scenario(scenario_name=None):
    from nebula.scenarios import ScenarioManagement

    if settings.advanced_analytics:
        from aim.sdk.repo import Repo

        # NEBULALOGGER START
        try:
            repo = Repo.from_path(f"{settings.log_dir}")
            list_tuples_participant_hash = get_run_hashes_scenario(scenario_name)
            hashes = [tuple[1] for tuple in list_tuples_participant_hash]
            logging.info(f"Removing statistics from {scenario_name}: {hashes}")
            success, remaining_runs = repo.delete_runs(hashes)
            if success:
                logging.info(f"Successfully deleted {len(hashes)} runs.")
            else:
                logging.info("Something went wrong while deleting runs.")
                logging.info(f"Remaining runs: {remaining_runs}")
        except Exception as e:
            logging.exception(f"Error removing statistics from {scenario_name}: {e}")
            pass
        # NEBULALOGGER END
    # Remove registered nodes and conditions
    nodes_registration.pop(scenario_name, None)
    remove_nodes_by_scenario_name(scenario_name)
    remove_scenario_by_name(scenario_name)
    remove_note(scenario_name)
    ScenarioManagement.remove_files_by_scenario(scenario_name)


@app.get("/nebula/dashboard/{scenario_name}/remove")
async def nebula_remove_scenario(scenario_name: str, request: Request, session: dict = Depends(get_session)):
    if "user" in session:
        if session["role"] == "demo":
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
        elif session["role"] == "user":
            if not check_scenario_with_role(session["role"], scenario_name):
                raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
        remove_scenario(scenario_name)
        return RedirectResponse(url="/nebula/dashboard")
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)


if settings.advanced_analytics:
    logging.info("Advanced analytics enabled")

    # NEBULALOGGER START
    def get_tracking_hash_scenario(scenario_name):
        import requests

        url = f"http://127.0.0.1:{settings.statistics_port}/nebula/statistics/api/experiments"
        # Get JSON data from the URL
        response = requests.get(url)
        if response.status_code == 200:
            experiments = response.json()
            for experiment in experiments:
                if experiment["name"] == scenario_name:
                    return experiment["id"]

        return None

    @app.get("/nebula/dashboard/statistics/", response_class=HTMLResponse)
    @app.get("/nebula/dashboard/{scenario_name}/statistics/", response_class=HTMLResponse)
    async def nebula_dashboard_statistics(request: Request, scenario_name: str = None):
        statistics_url = "/nebula/statistics/"
        if scenario_name is not None:
            experiment_hash = get_tracking_hash_scenario(scenario_name=scenario_name)
            statistics_url += f"experiments/{experiment_hash}/runs"

        return templates.TemplateResponse("statistics.html", {"request": request, "statistics_url": statistics_url})

    @app.get(
        "/nebula/dashboard/{scenario_name}/node/{hash}/metrics",
        response_class=HTMLResponse,
    )
    async def nebula_dashboard_node_metrics(request: Request, scenario_name: str, hash: str):
        statistics_url = f"/nebula/statistics/runs/{hash}/metrics"
        return templates.TemplateResponse("statistics.html", {"request": request, "statistics_url": statistics_url})

    @app.api_route("/nebula/statistics/", methods=["GET", "POST"])
    @app.api_route("/nebula/statistics/{path:path}", methods=["GET", "POST"])
    async def statistics_proxy(request: Request, path: str = None, session: dict = Depends(get_session)):
        if "user" in session:
            query_string = urlencode(request.query_params)

            url = f"http://127.0.0.1:{settings.statistics_port}/nebula/statistics"
            url = f"{url}{('/' + path) if path else '/'}" + (f"?{query_string}" if query_string else "")

            headers = {key: value for key, value in request.headers.items() if key.lower() != "host"}

            response = requests.request(
                method=request.method,
                url=url,
                headers=headers,
                data=await request.body(),
                cookies=request.cookies,
                allow_redirects=False,
            )

            excluded_headers = [
                "content-encoding",
                "content-length",
                "transfer-encoding",
                "connection",
            ]
            filtered_headers = [
                (name, value) for name, value in response.raw.headers.items() if name.lower() not in excluded_headers
            ]

            return Response(
                content=response.content,
                status_code=response.status_code,
                headers=dict(filtered_headers),
            )
        else:
            raise HTTPException(status_code=401)

    @app.get("/nebula/dashboard/{scenario_name}/download/metrics")
    async def nebula_dashboard_download_metrics(
        scenario_name: str, request: Request, session: dict = Depends(get_session)
    ):
        from aim.sdk.repo import Repo

        if "user" in session:
            # Obtener las métricas del escenario
            os.makedirs(
                Utils.check_path(settings.log_dir, os.path.join(scenario_name, "metrics")),
                exist_ok=True,
            )

            aim_repo = Repo.from_path("/nebula/nebula/app/logs")
            query = f"run.experiment == '{scenario_name}'"
            df = aim_repo.query_metrics(query).dataframe()

            hash_to_participant = {hash: participant for participant, hash in get_run_hashes_scenario(scenario_name)}
            df["participant"] = df["run.hash"].map(hash_to_participant)
            df.drop(
                columns=["run", "run.hash", "metric.context", "epoch"],
                axis=1,
                inplace=True,
            )
            cols = df.columns.tolist()
            cols.remove("participant")
            cols.remove("metric.name")
            df = df.reindex(columns=["participant", "metric.name"] + cols)

            for name, group in df.groupby("participant"):
                group.to_csv(
                    os.path.join(settings.log_dir, scenario_name, "metrics", f"{name}.csv"),
                    index=True,
                )

            # Crear un archivo zip con las métricas, enviarlo al usuario y eliminarlo
            memory_file = io.BytesIO()
            with zipfile.ZipFile(memory_file, "w", zipfile.ZIP_DEFLATED) as zipf:
                zipdir(os.path.join(settings.log_dir, scenario_name, "metrics"), zipf)

            memory_file.seek(0)

            return StreamingResponse(
                memory_file,
                media_type="application/zip",
                headers={"Content-Disposition": f"attachment; filename={scenario_name}_metrics.zip"},
            )
        else:
            raise HTTPException(status_code=401)

    # NEBULALOGGER END
else:
    logging.info("Advanced analytics disabled")

    # TENSORBOARD START
    @app.get("/nebula/dashboard/statistics/", response_class=HTMLResponse)
    @app.get("/nebula/dashboard/{scenario_name}/statistics/", response_class=HTMLResponse)
    async def nebula_dashboard_statistics(request: Request, scenario_name: str = None):
        statistics_url = "/nebula/statistics/"
        if scenario_name is not None:
            statistics_url += f"?smoothing=0&runFilter={scenario_name}"

        return templates.TemplateResponse("statistics.html", {"request": request, "statistics_url": statistics_url})

    @app.api_route("/nebula/statistics/", methods=["GET", "POST"])
    @app.api_route("/nebula/statistics/{path:path}", methods=["GET", "POST"])
    async def statistics_proxy(request: Request, path: str = None, session: dict = Depends(get_session)):
        if "user" in session:
            query_string = urlencode(request.query_params)

            url = "http://localhost:8080"
            tensorboard_url = f"{url}{('/' + path) if path else ''}" + ("?" + query_string if query_string else "")

            headers = {key: value for key, value in request.headers.items() if key.lower() != "host"}

            response = requests.request(
                method=request.method,
                url=tensorboard_url,
                headers=headers,
                data=await request.body(),
                cookies=request.cookies,
                allow_redirects=False,
            )

            excluded_headers = [
                "content-encoding",
                "content-length",
                "transfer-encoding",
                "connection",
            ]

            filtered_headers = [
                (name, value) for name, value in response.raw.headers.items() if name.lower() not in excluded_headers
            ]

            if "text/html" in response.headers["Content-Type"]:
                content = response.text
                content = content.replace("url(/", "url(/nebula/statistics/")
                content = content.replace('src="/', 'src="/nebula/statistics/')
                content = content.replace('href="/', 'href="/nebula/statistics/')
                response = Response(content, response.status_code, dict(filtered_headers))
                return response

            if path and path.endswith(".js"):
                content = response.text
                content = content.replace(
                    "experiment/${s}/data/plugin",
                    "nebula/statistics/experiment/${s}/data/plugin",
                )
                response = Response(content, response.status_code, dict(filtered_headers))
                return response

            return Response(response.content, response.status_code, dict(filtered_headers))

        else:
            raise HTTPException(status_code=401)

    @app.get("/experiment/{path:path}")
    @app.post("/experiment/{path:path}")
    async def metrics_proxy(path: str = None, request: Request = None):
        query_params = request.query_params
        new_url = "/nebula/statistics/experiment/" + path
        if query_params:
            new_url += "?" + urlencode(query_params)

        return RedirectResponse(url=new_url)

    # TENSORBOARD END


def zipdir(path, ziph):
    # ziph is zipfile handle
    for root, dirs, files in os.walk(path):
        for file in files:
            ziph.write(
                os.path.join(root, file),
                os.path.relpath(os.path.join(root, file), os.path.join(path, "..")),
            )


@app.get("/nebula/dashboard/{scenario_name}/download/logs")
async def nebula_dashboard_download_logs_metrics(
    scenario_name: str, request: Request, session: dict = Depends(get_session)
):
    if "user" in session:
        log_folder = Utils.check_path(settings.log_dir, scenario_name)
        config_folder = Utils.check_path(settings.config_dir, scenario_name)
        if os.path.exists(log_folder) and os.path.exists(config_folder):
            # Crear un archivo zip con los logs y los archivos de configuración, enviarlo al usuario
            memory_file = io.BytesIO()
            with zipfile.ZipFile(memory_file, "w", zipfile.ZIP_DEFLATED) as zipf:
                zipdir(log_folder, zipf)
                zipdir(config_folder, zipf)

            memory_file.seek(0)

            return StreamingResponse(
                memory_file,
                media_type="application/zip",
                headers={"Content-Disposition": f"attachment; filename={scenario_name}.zip"},
            )
        else:
            raise HTTPException(status_code=404, detail="Log or config folder not found")
    else:
        raise HTTPException(status_code=401)


@app.get("/nebula/dashboard/deployment/", response_class=HTMLResponse)
async def nebula_dashboard_deployment(request: Request, session: dict = Depends(get_session)):
    scenario_running = get_running_scenario()
    return templates.TemplateResponse(
        "deployment.html",
        {
            "request": request,
            "scenario_running": scenario_running,
            "user_logged_in": session.get("user"),
            "gpu_available": settings.gpu_available,
        },
    )


def attack_node_assign(
    nodes,
    federation,
    attack,
    poisoned_node_percent,
    poisoned_sample_percent,
    poisoned_noise_percent,
):
    """Identify which nodes will be attacked"""
    import math
    import random

    attack_matrix = []
    n_nodes = len(nodes)
    if n_nodes == 0:
        return attack_matrix

    nodes_index = []
    # Get the nodes index
    if federation == "DFL":
        nodes_index = list(nodes.keys())
    else:
        for node in nodes:
            if nodes[node]["role"] != "server":
                nodes_index.append(node)

    n_nodes = len(nodes_index)
    # Number of attacked nodes, round up
    num_attacked = int(math.ceil(poisoned_node_percent / 100 * n_nodes))
    if num_attacked > n_nodes:
        num_attacked = n_nodes

    # Get the index of attacked nodes
    attacked_nodes = random.sample(nodes_index, num_attacked)

    # Assign the role of each node
    for node in nodes:
        node_att = "No Attack"
        attack_sample_persent = 0
        poisoned_ratio = 0
        if (node in attacked_nodes) or (nodes[node]["malicious"]):
            node_att = attack
            attack_sample_persent = poisoned_sample_percent / 100
            poisoned_ratio = poisoned_noise_percent / 100
        nodes[node]["attacks"] = node_att
        nodes[node]["poisoned_sample_percent"] = attack_sample_persent
        nodes[node]["poisoned_ratio"] = poisoned_ratio
        attack_matrix.append([node, node_att, attack_sample_persent, poisoned_ratio])
    return nodes, attack_matrix


import math


def mobility_assign(nodes, mobile_participants_percent):
    """Assign mobility to nodes"""
    import random

    # Number of mobile nodes, round down
    num_mobile = math.floor(mobile_participants_percent / 100 * len(nodes))
    if num_mobile > len(nodes):
        num_mobile = len(nodes)

    # Get the index of mobile nodes
    mobile_nodes = random.sample(list(nodes.keys()), num_mobile)

    # Assign the role of each node
    for node in nodes:
        node_mob = False
        if node in mobile_nodes:
            node_mob = True
        nodes[node]["mobility"] = node_mob
    return nodes


# Stop all scenarios in the scenarios_list
stop_all_scenarios_event = asyncio.Event()

# Finish actual scenario
finish_scenario_event = asyncio.Event()

# Nodes that completed the experiment
nodes_finished = []


# Recieve a stopped node
@app.post("/nebula/dashboard/{scenario_name}/node/done")
async def node_stopped(scenario_name: str, request: Request):
    if request.headers.get("content-type") == "application/json":
        data = await request.json()
        nodes_finished.append(data["idx"])
        nodes_list = list_nodes_by_scenario_name(scenario_name)
        finished = True
        # Check if all the nodes of the scenario have finished the experiment
        for node in nodes_list:
            if str(node[1]) not in map(str, nodes_finished):
                finished = False

        if finished:
            nodes_finished.clear()
            finish_scenario_event.set()
            return JSONResponse(
                status_code=200,
                content={"message": "All nodes finished, scenario marked as completed."},
            )
        else:
            return JSONResponse(
                status_code=200,
                content={"message": "Node marked as finished, waiting for other nodes."},
            )
    else:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST)


async def run_scenario(scenario_data, role):
    import subprocess

    from nebula.scenarios import ScenarioManagement

    # Manager for the actual scenario
    scenarioManagement = ScenarioManagement(scenario_data)

    scenario_update_record(
        scenario_name=scenarioManagement.scenario_name,
        start_time=scenarioManagement.start_date_scenario,
        end_time="",
        status="running",
        title=scenario_data["scenario_title"],
        description=scenario_data["scenario_description"],
        network_subnet=scenario_data["network_subnet"],
        model=scenario_data["model"],
        dataset=scenario_data["dataset"],
        rounds=scenario_data["rounds"],
        role=role,
    )

    # Run the actual scenario
    try:
        if scenarioManagement.scenario.mobility:
            additional_participants = scenario_data["additional_participants"]
            schema_additional_participants = scenario_data["schema_additional_participants"]
            scenarioManagement.load_configurations_and_start_nodes(
                additional_participants, schema_additional_participants
            )
        else:
            scenarioManagement.load_configurations_and_start_nodes()
    except subprocess.CalledProcessError as e:
        logging.exception(f"Error docker-compose up: {e}")
        return

    nodes_registration[scenarioManagement.scenario_name] = {
        "n_nodes": scenario_data["n_nodes"],
        "nodes": set(),
    }

    nodes_registration[scenarioManagement.scenario_name]["condition"] = asyncio.Condition()

    return scenarioManagement.scenario_name


# Deploy the list of scenarios
async def run_scenarios(data, role):
    global scenarios_finished
    for scenario_data in data:
        finish_scenario_event.clear()
        logging.info(f"Running scenario {scenario_data['scenario_title']}")
        scenario_name = await run_scenario(scenario_data, role)
        # Waits till the scenario is completed
        while not finish_scenario_event.is_set() and not stop_all_scenarios_event.is_set():
            await asyncio.sleep(1)
        if stop_all_scenarios_event.is_set():
            stop_all_scenarios_event.clear()
            stop_scenario(scenario_name)
            return
        scenarios_finished = scenarios_finished + 1
        stop_scenario(scenario_name)
        await asyncio.sleep(5)


@app.post("/nebula/dashboard/deployment/run")
async def nebula_dashboard_deployment_run(
    request: Request,
    background_tasks: BackgroundTasks,
    session: dict = Depends(get_session),
):
    if "user" not in session or session["role"] in ["demo", "user"] and get_running_scenario():
        raise HTTPException(status_code=401)

    if request.headers.get("content-type") != "application/json":
        raise HTTPException(status_code=401)

    stop_all_scenarios()
    finish_scenario_event.clear()
    stop_all_scenarios_event.clear()
    data = await request.json()
    global scenarios_finished, scenarios_list_length
    scenarios_finished = 0
    scenarios_list_length = len(data)
    logging.info(f"Running deployment with {len(data)} scenarios")
    background_tasks.add_task(run_scenarios, data, session["role"])
    return RedirectResponse(url="/nebula/dashboard", status_code=303)
    # return Response(content="Success", status_code=200)


if __name__ == "__main__":
    # Parse args from command line
    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=int, default=5000, help="Port to run the frontend on.")
    args = parser.parse_args()
    logging.info(f"Starting frontend on port {args.port}")
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=args.port)



import logging
import os
import random

import numpy as np
import torch


def enable_deterministic(config):
    seed = config.participant["scenario_args"]["random_seed"]
    logging.info(f"Fixing randomness with seed {seed}")
    np.random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



import os
import shutil
import sys
import zipfile

import torch
from torchvision.datasets import MNIST, utils

from nebula.core.datasets.nebuladataset import NebulaDataset


class KITSUN(MNIST):
    def __init__(self, train=True):
        self.root = f"{sys.path[0]}/data"
        self.download = True
        self.train = train
        super(MNIST, self).__init__(self.root)
        self.training_file = f"{self.root}/kitsun/processed/kitsun_train.pt"
        self.test_file = f"{self.root}/kitsun/processed/kitsun_test.pt"

        if not os.path.exists(f"{self.root}/kitsun/processed/kitsun_test.pt") or not os.path.exists(
            f"{self.root}/kitsun/processed/kitsun_train.pt"
        ):
            if self.download:
                self.dataset_download()
            else:
                raise RuntimeError("Dataset not found, set parameter download=True to download")

        if self.train:
            data_file = self.training_file
        else:
            data_file = self.test_file
        data_and_targets = torch.load(data_file)
        self.data, self.targets = data_and_targets[0], data_and_targets[1]
        self.data = self.data

    def __getitem__(self, index):
        img, target = self.data[index], int(self.targets[index])
        if self.transform is not None:
            img = img
        if self.target_transform is not None:
            target = target
        return img, target

    def dataset_download(self):
        paths = [f"{self.root}/kitsun/raw/", f"{self.root}/kitsun/processed/"]
        for path in paths:
            if not os.path.exists(path):
                os.makedirs(path, exist_ok=True)
        data_link = "https://files.ifi.uzh.ch/CSG/research/fl/data/kitsun.zip"
        filename = data_link.split("/")[-1]

        utils.download_and_extract_archive(data_link, download_root=f"{self.root}/kitsun/raw/", filename=filename)

        with zipfile.ZipFile(f"{self.root}/kitsun/raw/{filename}", "r") as zip_ref:
            zip_ref.extractall(f"{self.root}/kitsun/raw/")

        train_raw = f"{self.root}/kitsun/raw/kitsun_train.pt"
        test_raw = f"{self.root}/kitsun/raw/kitsun_test.pt"
        train_file = f"{self.root}/kitsun/processed/kitsun_train.pt"
        test_file = f"{self.root}/kitsun/processed/kitsun_test.pt"
        if not os.path.exists(train_file):
            shutil.copy(train_raw, train_file)
        if not os.path.exists(test_file):
            shutil.copy(test_raw, test_file)


class KITSUNDataset(NebulaDataset):
    def __init__(
        self,
        num_classes=10,
        partition_id=0,
        partitions_number=1,
        batch_size=32,
        num_workers=4,
        iid=True,
        partition="dirichlet",
        partition_parameter=0.5,
        seed=42,
        config=None,
    ):
        super().__init__(
            num_classes=num_classes,
            partition_id=partition_id,
            partitions_number=partitions_number,
            batch_size=batch_size,
            num_workers=num_workers,
            iid=iid,
            partition=partition,
            partition_parameter=partition_parameter,
            seed=seed,
            config=config,
        )

    def initialize_dataset(self):
        # Load CIFAR10 train dataset
        if self.train_set is None:
            self.train_set = self.load_kitsun_dataset(train=True)
        if self.test_set is None:
            self.test_set = self.load_kitsun_dataset(train=False)

        # All nodes have the same test set (indices are the same for all nodes)
        self.test_indices_map = list(range(len(self.test_set)))

        # Depending on the iid flag, generate a non-iid or iid map of the train set
        if self.iid:
            self.train_indices_map = self.generate_iid_map(self.train_set, self.partition, self.partition_parameter)
            self.local_test_indices_map = self.generate_iid_map(self.test_set, self.partition, self.partition_parameter)
        else:
            self.train_indices_map = self.generate_non_iid_map(self.train_set, self.partition, self.partition_parameter)
            self.local_test_indices_map = self.generate_non_iid_map(
                self.test_set, self.partition, self.partition_parameter
            )

        print(f"Length of train indices map: {len(self.train_indices_map)}")
        print(f"Lenght of test indices map (global): {len(self.test_indices_map)}")
        print(f"Length of test indices map (local): {len(self.local_test_indices_map)}")

    def load_kitsun_dataset(self, train=True):
        if train:
            return KITSUN(train=True)
        return KITSUN(train=False)

    def generate_non_iid_map(self, dataset, partition="dirichlet", partition_parameter=0.5):
        if partition == "dirichlet":
            partitions_map = self.dirichlet_partition(dataset, alpha=partition_parameter)
        elif partition == "percent":
            partitions_map = self.percentage_partition(dataset, percentage=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for Non-IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]

    def generate_iid_map(self, dataset, partition="balancediid", partition_parameter=2):
        if partition == "balancediid":
            partitions_map = self.balanced_iid_partition(dataset)
        elif partition == "unbalancediid":
            partitions_map = self.unbalanced_iid_partition(dataset, imbalance_factor=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]



import ast
import os
import sys
import zipfile

import pandas as pd
import torch
from sklearn.model_selection import train_test_split
from torchvision.datasets import MNIST, utils

from nebula.core.datasets.nebuladataset import NebulaDataset


class SYSCALL(MNIST):
    def __init__(
        self,
        partition_id,
        partitions_number,
        root_dir,
        train=True,
        transform=None,
        target_transform=None,
        download=False,
    ):
        super().__init__(root_dir, transform=None, target_transform=None)
        self.transform = transform
        self.target_transform = target_transform
        self.partition_id = partition_id
        self.partitions_number = partitions_number
        self.download = download
        self.download_link = "https://files.ifi.uzh.ch/CSG/research/fl/data/syscall.zip"
        self.train = train
        self.root = root_dir
        self.training_file = f"{self.root}/syscall/processed/syscall_train.pt"
        self.test_file = f"{self.root}/syscall/processed/syscall_test.pt"

        if not os.path.exists(f"{self.root}/syscall/processed/syscall_test.pt") or not os.path.exists(
            f"{self.root}/syscall/processed/syscall_train.pt"
        ):
            if self.download:
                self.dataset_download()
                self.process()
            else:
                raise RuntimeError("Dataset not found, set parameter download=True to download")
        else:
            print("SYSCALL dataset already downloaded and processed.")

        if self.train:
            data_file = self.training_file
        else:
            data_file = self.test_file
        data_and_targets = torch.load(data_file)
        self.data, self.targets = data_and_targets[0], data_and_targets[1]

    def __getitem__(self, index):
        img, target = self.data[index], int(self.targets[index])
        if self.transform is not None:
            img = img
        if self.target_transform is not None:
            target = target
        return img, target

    def dataset_download(self):
        paths = [f"{self.root}/syscall/raw/", f"{self.root}/syscall/processed/"]
        for path in paths:
            if not os.path.exists(path):
                os.makedirs(path, exist_ok=True)
        print("Downloading SYSCALL dataset...")
        filename = self.download_link.split("/")[-1]
        utils.download_and_extract_archive(
            self.download_link,
            download_root=f"{self.root}/syscall/raw/",
            filename=filename,
        )

        with zipfile.ZipFile(f"{self.root}/syscall/raw/{filename}", "r") as zip_ref:
            zip_ref.extractall(f"{self.root}/syscall/raw/")

    def process(self):
        print("Processing SYSCALL dataset...")
        df = pd.DataFrame()
        files = os.listdir(f"{self.root}/syscall/raw/")
        feature_name = "system calls frequency_1gram-scaled"
        for f in files:
            if ".csv" in f:
                fi_path = f"{self.root}/syscall/raw/{f}"
                csv_df = pd.read_csv(fi_path, sep="\t")
                feature = [ast.literal_eval(i) for i in csv_df[feature_name]]
                csv_df[feature_name] = feature
                df = pd.concat([df, csv_df])
        df["maltype"] = df["maltype"].replace(to_replace="normalv2", value="normal")
        classes_to_targets = {}
        t = 0
        for i in set(df["maltype"]):
            classes_to_targets[i] = t
            t += 1
        classes = list(classes_to_targets.keys())

        for c in classes_to_targets:
            df["maltype"] = df["maltype"].replace(to_replace=c, value=classes_to_targets[c])

        all_targes = torch.tensor(df["maltype"].tolist())
        all_data = torch.tensor(df[feature_name].tolist())

        x_train, x_test, y_train, y_test = train_test_split(all_data, all_targes, test_size=0.15, random_state=42)
        train = [x_train, y_train, classes_to_targets, classes]
        test = [x_test, y_test, classes_to_targets, classes]
        train_file = f"{self.root}/syscall/processed/syscall_train.pt"
        test_file = f"{self.root}/syscall/processed/syscall_test.pt"
        if not os.path.exists(train_file):
            torch.save(train, train_file)
        if not os.path.exists(test_file):
            torch.save(test, test_file)


class SYSCALLDataset(NebulaDataset):
    def __init__(
        self,
        num_classes=10,
        partition_id=0,
        partitions_number=1,
        batch_size=32,
        num_workers=4,
        iid=True,
        partition="dirichlet",
        partition_parameter=0.5,
        seed=42,
        config=None,
    ):
        super().__init__(
            num_classes=num_classes,
            partition_id=partition_id,
            partitions_number=partitions_number,
            batch_size=batch_size,
            num_workers=num_workers,
            iid=iid,
            partition=partition,
            partition_parameter=partition_parameter,
            seed=seed,
            config=config,
        )

    def initialize_dataset(self):
        # Load syscall train dataset
        if self.train_set is None:
            self.train_set = self.load_syscall_dataset(train=True)
        if self.test_set is None:
            self.test_set = self.load_syscall_dataset(train=False)

        # All nodes have the same test set (indices are the same for all nodes)
        self.test_indices_map = list(range(len(self.test_set)))

        # Depending on the iid flag, generate a non-iid or iid map of the train set
        if self.iid:
            self.train_indices_map = self.generate_iid_map(self.train_set, self.partition, self.partition_parameter)
            self.local_test_indices_map = self.generate_iid_map(self.test_set, self.partition, self.partition_parameter)
        else:
            self.train_indices_map = self.generate_non_iid_map(self.train_set, self.partition, self.partition_parameter)
            self.local_test_indices_map = self.generate_non_iid_map(
                self.test_set, self.partition, self.partition_parameter
            )

        print(f"Length of train indices map: {len(self.train_indices_map)}")
        print(f"Lenght of test indices map: {len(self.test_indices_map)}")

    def load_syscall_dataset(self, train=True):
        if train:
            return SYSCALL(
                partition_id=self.partition_id,
                partitions_number=self.partitions_number,
                root_dir=f"{sys.path[0]}/data",
                train=True,
            )
        else:
            return SYSCALL(
                partition_id=self.partition_id,
                partitions_number=self.partitions_number,
                root_dir=f"{sys.path[0]}/data",
                train=False,
            )

    def generate_non_iid_map(self, dataset, partition="dirichlet", partition_parameter=0.5):
        if partition == "dirichlet":
            partitions_map = self.dirichlet_partition(dataset, alpha=partition_parameter)
        elif partition == "percent":
            partitions_map = self.percentage_partition(dataset, percentage=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for Non-IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]

    def generate_iid_map(self, dataset, partition="balancediid", partition_parameter=2):
        if partition == "balancediid":
            partitions_map = self.balanced_iid_partition(dataset)
        elif partition == "unbalancediid":
            partitions_map = self.unbalanced_iid_partition(dataset, imbalance_factor=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]



import logging
import pickle
import traceback

from sklearn.metrics import accuracy_score


class Scikit:
    def __init__(self, model, data, config=None, logger=None):
        self.model = model
        self.data = data
        self.config = config
        self.logger = logger
        self.round = 0
        self.epochs = 1
        self.logger.log_data({"Round": self.round}, step=self.logger.global_step)

    def set_model(self, model):
        self.model = model

    def get_round(self):
        return self.round

    def set_data(self, data):
        self.data = data

    def serialize_model(self, params=None):
        if params is None:
            params = self.model.get_params()
        return pickle.dumps(params)

    def deserialize_model(self, data):
        try:
            params = pickle.loads(data)
            return params
        except:
            raise Exception("Error decoding parameters")

    def set_model_parameters(self, params):
        self.model.set_params(**params)

    def get_model_parameters(self):
        return self.model.get_params()

    def set_epochs(self, epochs):
        self.epochs = epochs

    def fit(self):
        try:
            X_train, y_train = self.data.train_dataloader()
            self.model.fit(X_train, y_train)
        except Exception as e:
            logging.exception(f"Error with scikit-learn fit. {e}")
            logging.exception(traceback.format_exc())

    def interrupt_fit(self):
        pass

    def evaluate(self):
        try:
            X_test, y_test = self.data.test_dataloader()
            y_pred = self.model.predict(X_test)
            accuracy = accuracy_score(y_test, y_pred)
            logging.info(f"Accuracy: {accuracy}")
        except Exception as e:
            logging.exception(f"Error with scikit-learn evaluate. {e}")
            logging.exception(traceback.format_exc())
            return None

    def get_train_size(self):
        return (
            len(self.data.train_dataloader()),
            len(self.data.test_dataloader()),
        )

    def finalize_round(self):
        self.round += 1
        if self.logger:
            self.logger.log_data({"Round": self.round})



import asyncio
import inspect
import logging
from collections import defaultdict
from functools import wraps


def event_handler(message_type, action):
    """Decorator for registering an event handler."""

    def decorator(func):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            return await func(*args, **kwargs)

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        if asyncio.iscoroutinefunction(func):
            wrapper = async_wrapper
        else:
            wrapper = sync_wrapper

        action_name = message_type.Action.Name(action) if action is not None else "None"
        wrapper._event_handler = (message_type.DESCRIPTOR.full_name, action_name)
        return wrapper

    return decorator


class EventManager:
    def __init__(self, default_callbacks=None):
        self._event_callbacks = defaultdict(list)
        self._register_default_callbacks(default_callbacks or [])

    def _register_default_callbacks(self, default_callbacks):
        """Registers default callbacks for events."""
        for callback in default_callbacks:
            handler_info = getattr(callback, "_event_handler", None)
            if handler_info is not None:
                self.register_event(handler_info, callback)
            else:
                raise ValueError("The callback must be decorated with @event_handler.")

    def register_event(self, handler_info, callback):
        """Records a callback for a specific event."""
        if callable(callback):
            self._event_callbacks[handler_info].append(callback)
        else:
            raise ValueError("The callback must be a callable function.")

    def unregister_event(self, handler_info, callback):
        """Unregisters a previously registered callback for an event."""
        if callback in self._event_callbacks[handler_info]:
            self._event_callbacks[handler_info].remove(callback)

    async def trigger_event(self, source, message, *args, **kwargs):
        """Triggers an event, executing all associated callbacks."""
        message_type = message.DESCRIPTOR.full_name
        if hasattr(message, "action"):
            action_name = message.Action.Name(message.action)
        else:
            action_name = "None"

        handler_info = (message_type, action_name)

        if handler_info in self._event_callbacks:
            for callback in self._event_callbacks[handler_info]:
                try:
                    if asyncio.iscoroutinefunction(callback) or inspect.iscoroutine(callback):
                        await callback(source, message, *args, **kwargs)
                    else:
                        callback(source, message, *args, **kwargs)
                except Exception as e:
                    logging.exception(f"Error executing callback for {handler_info}: {e}")
        else:
            logging.error(f"No callbacks registered for event {handler_info}")

    async def get_event_callbacks(self, event_name):
        """Returns the callbacks for a specific event."""
        return self._event_callbacks[event_name]

    def get_event_callbacks_names(self):
        """Returns the names of the registered events."""
        return self._event_callbacks.keys()



import glob
import hashlib
import json
import logging
import math
import os
import shutil
import subprocess
import sys
import textwrap
import time
from datetime import datetime

import docker

from nebula.addons.blockchain.blockchain_deployer import BlockchainDeployer
from nebula.addons.topologymanager import TopologyManager
from nebula.config.config import Config
from nebula.core.utils.certificate import generate_ca_certificate, generate_certificate
from nebula.frontend.utils import Utils


# Definition of a scenario
class Scenario:
    def __init__(
        self,
        scenario_title,
        scenario_description,
        deployment,
        federation,
        topology,
        nodes,
        nodes_graph,
        n_nodes,
        matrix,
        dataset,
        iid,
        partition_selection,
        partition_parameter,
        model,
        agg_algorithm,
        rounds,
        logginglevel,
        accelerator,
        network_subnet,
        network_gateway,
        epochs,
        attacks,
        poisoned_node_percent,
        poisoned_sample_percent,
        poisoned_noise_percent,
        with_reputation,
        is_dynamic_topology,
        is_dynamic_aggregation,
        target_aggregation,
        random_geo,
        latitude,
        longitude,
        mobility,
        mobility_type,
        radius_federation,
        scheme_mobility,
        round_frequency,
        mobile_participants_percent,
        additional_participants,
        schema_additional_participants,
    ):
        """
        Initialize the scenario.

        Args:
            scenario_title (str): Title of the scenario.
            scenario_description (str): Description of the scenario.
            deployment (str): Type of deployment (e.g., 'docker', 'process').
            federation (str): Type of federation.
            topology (str): Network topology.
            nodes (dict): Dictionary of nodes.
            nodes_graph (dict): Graph of nodes.
            n_nodes (int): Number of nodes.
            matrix (list): Matrix of connections.
            dataset (str): Dataset used.
            iid (bool): Indicator if data is independent and identically distributed.
            partition_selection (str): Method of partition selection.
            partition_parameter (float): Parameter for partition selection.
            model (str): Model used.
            agg_algorithm (str): Aggregation algorithm.
            rounds (int): Number of rounds.
            logginglevel (str): Logging level.
            accelerator (str): Accelerator used.
            network_subnet (str): Network subnet.
            network_gateway (str): Network gateway.
            epochs (int): Number of epochs.
            attacks (list): List of attacks.
            poisoned_node_percent (float): Percentage of poisoned nodes.
            poisoned_sample_percent (float): Percentage of poisoned samples.
            is_dynamic_topology (bool): Indicator if topology is dynamic.
            is_dynamic_aggregation (bool): Indicator if aggregation is dynamic.
            target_aggregation (str): Target aggregation method.
            random_geo (bool): Indicator if random geo is used.
            latitude (float): Latitude for mobility.
            longitude (float): Longitude for mobility.
            mobility (bool): Indicator if mobility is used.
            mobility_type (str): Type of mobility.
            radius_federation (float): Radius of federation.
            scheme_mobility (str): Scheme of mobility.
            round_frequency (int): Frequency of rounds.
            mobile_participants_percent (float): Percentage of mobile participants.
            additional_participants (list): List of additional participants.
            schema_additional_participants (str): Schema for additional participants.
        """
        self.scenario_title = scenario_title
        self.scenario_description = scenario_description
        self.deployment = deployment
        self.federation = federation
        self.topology = topology
        self.nodes = nodes
        self.nodes_graph = nodes_graph
        self.n_nodes = n_nodes
        self.matrix = matrix
        self.dataset = dataset
        self.iid = iid
        self.partition_selection = partition_selection
        self.partition_parameter = partition_parameter
        self.model = model
        self.agg_algorithm = agg_algorithm
        self.rounds = rounds
        self.logginglevel = logginglevel
        self.accelerator = accelerator
        self.network_subnet = network_subnet
        self.network_gateway = network_gateway
        self.epochs = epochs
        self.attacks = attacks
        self.poisoned_node_percent = poisoned_node_percent
        self.poisoned_sample_percent = poisoned_sample_percent
        self.poisoned_noise_percent = poisoned_noise_percent
        self.with_reputation = with_reputation
        self.is_dynamic_topology = is_dynamic_topology
        self.is_dynamic_aggregation = is_dynamic_aggregation
        self.target_aggregation = target_aggregation
        self.random_geo = random_geo
        self.latitude = latitude
        self.longitude = longitude
        self.mobility = mobility
        self.mobility_type = mobility_type
        self.radius_federation = radius_federation
        self.scheme_mobility = scheme_mobility
        self.round_frequency = round_frequency
        self.mobile_participants_percent = mobile_participants_percent
        self.additional_participants = additional_participants
        self.schema_additional_participants = schema_additional_participants

    def attack_node_assign(
        self,
        nodes,
        federation,
        attack,
        poisoned_node_percent,
        poisoned_sample_percent,
        poisoned_noise_percent,
    ):
        """Identify which nodes will be attacked"""
        import math
        import random

        nodes_index = []
        # Get the nodes index
        if federation == "DFL":
            nodes_index = list(nodes.keys())
        else:
            for node in nodes:
                if nodes[node]["role"] != "server":
                    nodes_index.append(node)

        mal_nodes_defined = any(nodes[node]["malicious"] for node in nodes)

        attacked_nodes = []

        if not mal_nodes_defined:
            n_nodes = len(nodes_index)
            # Number of attacked nodes, round up
            num_attacked = int(math.ceil(poisoned_node_percent / 100 * n_nodes))
            if num_attacked > n_nodes:
                num_attacked = n_nodes

            # Get the index of attacked nodes
            attacked_nodes = random.sample(nodes_index, num_attacked)

        # Assign the role of each node
        for node in nodes:
            node_att = "No Attack"
            malicious = False
            attack_sample_percent = 0
            poisoned_ratio = 0
            if (str(nodes[node]["id"]) in attacked_nodes) or (nodes[node]["malicious"]):
                malicious = True
                node_att = attack
                attack_sample_percent = poisoned_sample_percent / 100
                poisoned_ratio = poisoned_noise_percent / 100
            nodes[node]["malicious"] = malicious
            nodes[node]["attacks"] = node_att
            nodes[node]["poisoned_sample_percent"] = attack_sample_percent
            nodes[node]["poisoned_ratio"] = poisoned_ratio
        return nodes

    def mobility_assign(self, nodes, mobile_participants_percent):
        """Assign mobility to nodes"""
        import random

        # Number of mobile nodes, round down
        num_mobile = math.floor(mobile_participants_percent / 100 * len(nodes))
        if num_mobile > len(nodes):
            num_mobile = len(nodes)

        # Get the index of mobile nodes
        mobile_nodes = random.sample(list(nodes.keys()), num_mobile)

        # Assign the role of each node
        for node in nodes:
            node_mob = False
            if node in mobile_nodes:
                node_mob = True
            nodes[node]["mobility"] = node_mob
        return nodes

    @classmethod
    def from_dict(cls, data):
        return cls(**data)


# Class to manage the current scenario
class ScenarioManagement:
    def __init__(self, scenario):
        # Current scenario
        self.scenario = Scenario.from_dict(scenario)
        # Scenario management settings
        self.start_date_scenario = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        self.scenario_name = f"nebula_{self.scenario.federation}_{datetime.now().strftime('%d_%m_%Y_%H_%M_%S')}"
        self.root_path = os.environ.get("NEBULA_ROOT_HOST")
        self.host_platform = os.environ.get("NEBULA_HOST_PLATFORM")
        self.config_dir = os.path.join(os.environ.get("NEBULA_CONFIG_DIR"), self.scenario_name)
        self.log_dir = os.environ.get("NEBULA_LOGS_DIR")
        self.cert_dir = os.environ.get("NEBULA_CERTS_DIR")
        self.advanced_analytics = os.environ.get("NEBULA_ADVANCED_ANALYTICS", "False") == "True"
        self.config = Config(entity="scenarioManagement")

        # Assign the controller endpoint
        if self.scenario.deployment == "docker":
            self.controller = "nebula-frontend"
        else:
            self.controller = f"127.0.0.1:{os.environ.get('NEBULA_FRONTEND_PORT')}"

        self.topologymanager = None
        self.env_path = None
        self.use_blockchain = self.scenario.agg_algorithm == "BlockchainReputation"

        # Create Scenario management dirs
        os.makedirs(self.config_dir, exist_ok=True)
        os.makedirs(os.path.join(self.log_dir, self.scenario_name), exist_ok=True)
        os.makedirs(self.cert_dir, exist_ok=True)

        # Give permissions to the directories
        os.chmod(self.config_dir, 0o777)
        os.chmod(os.path.join(self.log_dir, self.scenario_name), 0o777)
        os.chmod(self.cert_dir, 0o777)

        # Save the scenario configuration
        scenario_file = os.path.join(self.config_dir, "scenario.json")
        with open(scenario_file, "w") as f:
            json.dump(scenario, f, sort_keys=False, indent=2)

        os.chmod(scenario_file, 0o777)

        # Save management settings
        settings = {
            "scenario_name": self.scenario_name,
            "root_path": self.root_path,
            "config_dir": self.config_dir,
            "log_dir": self.log_dir,
            "cert_dir": self.cert_dir,
            "env": None,
            "use_blockchain": self.use_blockchain,
        }

        settings_file = os.path.join(self.config_dir, "settings.json")
        with open(settings_file, "w") as f:
            json.dump(settings, f, sort_keys=False, indent=2)

        os.chmod(settings_file, 0o777)

        self.scenario.nodes = self.scenario.attack_node_assign(
            self.scenario.nodes,
            self.scenario.federation,
            self.scenario.attacks,
            int(self.scenario.poisoned_node_percent),
            int(self.scenario.poisoned_sample_percent),
            int(self.scenario.poisoned_noise_percent),
        )

        if self.scenario.mobility:
            mobile_participants_percent = int(self.scenario.mobile_participants_percent)
            self.scenario.nodes = self.scenario.mobility_assign(self.scenario.nodes, mobile_participants_percent)
        else:
            self.scenario.nodes = self.scenario.mobility_assign(self.scenario.nodes, 0)

        # Save node settings
        for node in self.scenario.nodes:
            node_config = self.scenario.nodes[node]
            participant_file = os.path.join(self.config_dir, f"participant_{node_config['id']}.json")
            os.makedirs(os.path.dirname(participant_file), exist_ok=True)
            shutil.copy(
                os.path.join(
                    os.path.dirname(__file__),
                    "./frontend/config/participant.json.example",
                ),
                participant_file,
            )
            os.chmod(participant_file, 0o777)
            with open(participant_file) as f:
                participant_config = json.load(f)

            participant_config["network_args"]["ip"] = node_config["ip"]
            participant_config["network_args"]["port"] = int(node_config["port"])
            participant_config["device_args"]["idx"] = node_config["id"]
            participant_config["device_args"]["start"] = node_config["start"]
            participant_config["device_args"]["role"] = node_config["role"]
            participant_config["device_args"]["proxy"] = node_config["proxy"]
            participant_config["device_args"]["malicious"] = node_config["malicious"]
            participant_config["scenario_args"]["rounds"] = int(self.scenario.rounds)
            participant_config["data_args"]["dataset"] = self.scenario.dataset
            participant_config["data_args"]["iid"] = self.scenario.iid
            participant_config["data_args"]["partition_selection"] = self.scenario.partition_selection
            participant_config["data_args"]["partition_parameter"] = self.scenario.partition_parameter
            participant_config["model_args"]["model"] = self.scenario.model
            participant_config["training_args"]["epochs"] = int(self.scenario.epochs)
            participant_config["device_args"]["accelerator"] = self.scenario.accelerator
            participant_config["device_args"]["logging"] = self.scenario.logginglevel
            participant_config["aggregator_args"]["algorithm"] = self.scenario.agg_algorithm
            participant_config["adversarial_args"]["attacks"] = node_config["attacks"]
            participant_config["adversarial_args"]["poisoned_sample_percent"] = node_config["poisoned_sample_percent"]
            participant_config["adversarial_args"]["poisoned_ratio"] = node_config["poisoned_ratio"]
            participant_config["defense_args"]["with_reputation"] = self.scenario.with_reputation
            participant_config["defense_args"]["is_dynamic_topology"] = self.scenario.is_dynamic_topology
            participant_config["defense_args"]["is_dynamic_aggregation"] = self.scenario.is_dynamic_aggregation
            participant_config["defense_args"]["target_aggregation"] = self.scenario.target_aggregation
            participant_config["mobility_args"]["random_geo"] = self.scenario.random_geo
            participant_config["mobility_args"]["latitude"] = self.scenario.latitude
            participant_config["mobility_args"]["longitude"] = self.scenario.longitude
            participant_config["mobility_args"]["mobility"] = node_config["mobility"]
            participant_config["mobility_args"]["mobility_type"] = self.scenario.mobility_type
            participant_config["mobility_args"]["radius_federation"] = self.scenario.radius_federation
            participant_config["mobility_args"]["scheme_mobility"] = self.scenario.scheme_mobility
            participant_config["mobility_args"]["round_frequency"] = self.scenario.round_frequency

            with open(participant_file, "w") as f:
                json.dump(participant_config, f, sort_keys=False, indent=2)

    @staticmethod
    def stop_blockchain():
        if sys.platform == "win32":
            try:
                # Comando adaptado para PowerShell en Windows
                command = "docker ps -a --filter 'label=com.docker.compose.project=blockchain' --format '{{.ID}}' | ForEach-Object { docker rm --force --volumes $_ } | Out-Null"
                os.system(f'powershell.exe -Command "{command}"')
            except Exception as e:
                logging.exception(f"Error while killing docker containers: {e}")
        else:
            try:
                process = subprocess.Popen(
                    "docker ps -a --filter 'label=com.docker.compose.project=blockchain' --format '{{.ID}}' | xargs -n 1 docker rm --force --volumes  >/dev/null 2>&1",
                    shell=True,
                )
                process.wait()
            except subprocess.CalledProcessError:
                logging.exception("Docker Compose failed to stop blockchain or blockchain already exited.")

    @staticmethod
    def stop_participants():
        # When stopping the nodes, we need to remove the current_scenario_commands.sh file -> it will cause the nodes to stop using PIDs
        try:
            nebula_config_dir = os.environ.get("NEBULA_CONFIG_DIR")
            if not nebula_config_dir:
                current_dir = os.path.dirname(__file__)
                nebula_base_dir = os.path.abspath(os.path.join(current_dir, ".."))
                nebula_config_dir = os.path.join(nebula_base_dir, "app", "config")
                logging.info(f"NEBULA_CONFIG_DIR not found. Using default path: {nebula_config_dir}")
            if os.environ.get("NEBULA_HOST_PLATFORM") == "windows":
                scenario_commands_file = os.path.join(nebula_config_dir, "current_scenario_commands.ps1")
            else:
                scenario_commands_file = os.path.join(nebula_config_dir, "current_scenario_commands.sh")
            if os.path.exists(scenario_commands_file):
                os.remove(scenario_commands_file)
        except Exception as e:
            logging.exception(f"Error while removing current_scenario_commands.sh file: {e}")

        if sys.platform == "win32":
            try:
                # kill all the docker containers which contain the word "nebula-core"
                commands = [
                    """docker kill $(docker ps -q --filter ancestor=nebula-core) | Out-Null""",
                    """docker rm $(docker ps -a -q --filter ancestor=nebula-core) | Out-Null""",
                    r"""docker network rm $(docker network ls | Where-Object { ($_ -split '\s+')[1] -like 'nebula-net-scenario' } | ForEach-Object { ($_ -split '\s+')[0] }) | Out-Null""",
                ]

                for command in commands:
                    time.sleep(1)
                    exit_code = os.system(f'powershell.exe -Command "{command}"')
                    # logging.info(f"Windows Command '{command}' executed with exit code: {exit_code}")

            except Exception as e:
                raise Exception(f"Error while killing docker containers: {e}")
        else:
            try:
                commands = [
                    """docker kill $(docker ps -q --filter ancestor=nebula-core) > /dev/null 2>&1""",
                    """docker rm $(docker ps -a -q --filter ancestor=nebula-core) > /dev/null 2>&1""",
                    """docker network rm $(docker network ls | grep nebula-net-scenario | awk '{print $1}') > /dev/null 2>&1""",
                ]

                for command in commands:
                    time.sleep(1)
                    exit_code = os.system(command)
                    # logging.info(f"Linux Command '{command}' executed with exit code: {exit_code}")

            except Exception as e:
                raise Exception(f"Error while killing docker containers: {e}")

    @staticmethod
    def stop_nodes():
        logging.info("Closing NEBULA nodes... Please wait")
        ScenarioManagement.stop_participants()
        ScenarioManagement.stop_blockchain()

    def load_configurations_and_start_nodes(self, additional_participants=None, schema_additional_participants=None):
        logging.info(f"Generating the scenario {self.scenario_name} at {self.start_date_scenario}")

        # Generate CA certificate
        generate_ca_certificate(dir_path=self.cert_dir)

        # Get participants configurations
        participant_files = glob.glob(f"{self.config_dir}/participant_*.json")
        participant_files.sort()
        if len(participant_files) == 0:
            raise ValueError("No participant files found in config folder")

        self.config.set_participants_config(participant_files)
        self.n_nodes = len(participant_files)
        logging.info(f"Number of nodes: {self.n_nodes}")

        self.topologymanager = (
            self.create_topology(matrix=self.scenario.matrix) if self.scenario.matrix else self.create_topology()
        )

        # Update participants configuration
        is_start_node = False
        config_participants = []
        for i in range(self.n_nodes):
            with open(f"{self.config_dir}/participant_" + str(i) + ".json") as f:
                participant_config = json.load(f)
            participant_config["scenario_args"]["federation"] = self.scenario.federation
            participant_config["scenario_args"]["n_nodes"] = self.n_nodes
            participant_config["network_args"]["neighbors"] = self.topologymanager.get_neighbors_string(i)
            participant_config["scenario_args"]["name"] = self.scenario_name
            participant_config["scenario_args"]["start_time"] = self.start_date_scenario
            participant_config["device_args"]["idx"] = i
            participant_config["device_args"]["uid"] = hashlib.sha1(
                (
                    str(participant_config["network_args"]["ip"])
                    + str(participant_config["network_args"]["port"])
                    + str(self.scenario_name)
                ).encode()
            ).hexdigest()
            if participant_config["mobility_args"]["random_geo"]:
                (
                    participant_config["mobility_args"]["latitude"],
                    participant_config["mobility_args"]["longitude"],
                ) = TopologyManager.get_coordinates(random_geo=True)
            # If not, use the given coordinates in the frontend
            participant_config["tracking_args"]["local_tracking"] = "advanced" if self.advanced_analytics else "basic"
            participant_config["tracking_args"]["log_dir"] = self.log_dir
            participant_config["tracking_args"]["config_dir"] = self.config_dir

            # Generate node certificate
            keyfile_path, certificate_path = generate_certificate(
                dir_path=self.cert_dir,
                node_id=f"participant_{i}",
                ip=participant_config["network_args"]["ip"],
            )

            participant_config["security_args"]["certfile"] = certificate_path
            participant_config["security_args"]["keyfile"] = keyfile_path

            if participant_config["device_args"]["start"]:
                if not is_start_node:
                    is_start_node = True
                else:
                    raise ValueError("Only one node can be start node")
            with open(f"{self.config_dir}/participant_" + str(i) + ".json", "w") as f:
                json.dump(participant_config, f, sort_keys=False, indent=2)

            config_participants.append((
                participant_config["network_args"]["ip"],
                participant_config["network_args"]["port"],
                participant_config["device_args"]["role"],
            ))
        if not is_start_node:
            raise ValueError("No start node found")
        self.config.set_participants_config(participant_files)

        # Add role to the topology (visualization purposes)
        self.topologymanager.update_nodes(config_participants)
        self.topologymanager.draw_graph(path=f"{self.log_dir}/{self.scenario_name}/topology.png", plot=False)

        # Include additional participants (if any) as copies of the last participant
        additional_participants_files = []
        if additional_participants:
            last_participant_file = participant_files[-1]
            last_participant_index = len(participant_files)

            for i, additional_participant in enumerate(additional_participants):
                additional_participant_file = f"{self.config_dir}/participant_{last_participant_index + i}.json"
                shutil.copy(last_participant_file, additional_participant_file)

                with open(additional_participant_file) as f:
                    participant_config = json.load(f)

                participant_config["scenario_args"]["n_nodes"] = self.n_nodes + i + 1
                participant_config["device_args"]["idx"] = last_participant_index + i
                participant_config["network_args"]["neighbors"] = ""
                participant_config["network_args"]["ip"] = (
                    participant_config["network_args"]["ip"].rsplit(".", 1)[0]
                    + "."
                    + str(int(participant_config["network_args"]["ip"].rsplit(".", 1)[1]) + 1)
                )
                participant_config["device_args"]["uid"] = hashlib.sha1(
                    (
                        str(participant_config["network_args"]["ip"])
                        + str(participant_config["network_args"]["port"])
                        + str(self.scenario_name)
                    ).encode()
                ).hexdigest()
                participant_config["mobility_args"]["additional_node"]["status"] = True
                participant_config["mobility_args"]["additional_node"]["round_start"] = additional_participant["round"]

                with open(additional_participant_file, "w") as f:
                    json.dump(participant_config, f, sort_keys=False, indent=2)

                additional_participants_files.append(additional_participant_file)

        if additional_participants_files:
            self.config.add_participants_config(additional_participants_files)

        if self.scenario.deployment in ["docker", "process"]:
            if self.use_blockchain:
                self.start_blockchain()
            if self.scenario.deployment == "docker":
                self.start_nodes_docker()
            else:
                self.start_nodes_process()
        else:
            logging.info(
                f"Virtualization mode is disabled for scenario '{self.scenario_name}' with {self.n_nodes} nodes. Waiting for nodes to start manually..."
            )

    def create_topology(self, matrix=None):
        import numpy as np

        if matrix is not None:
            if self.n_nodes > 2:
                topologymanager = TopologyManager(
                    topology=np.array(matrix),
                    scenario_name=self.scenario_name,
                    n_nodes=self.n_nodes,
                    b_symmetric=True,
                    undirected_neighbor_num=self.n_nodes - 1,
                )
            else:
                topologymanager = TopologyManager(
                    topology=np.array(matrix),
                    scenario_name=self.scenario_name,
                    n_nodes=self.n_nodes,
                    b_symmetric=True,
                    undirected_neighbor_num=2,
                )
        elif self.scenario.topology == "fully":
            # Create a fully connected network
            topologymanager = TopologyManager(
                scenario_name=self.scenario_name,
                n_nodes=self.n_nodes,
                b_symmetric=True,
                undirected_neighbor_num=self.n_nodes - 1,
            )
            topologymanager.generate_topology()
        elif self.scenario.topology == "ring":
            # Create a partially connected network (ring-structured network)
            topologymanager = TopologyManager(scenario_name=self.scenario_name, n_nodes=self.n_nodes, b_symmetric=True)
            topologymanager.generate_ring_topology(increase_convergence=True)
        elif self.scenario.topology == "random":
            # Create network topology using topology manager (random)
            topologymanager = TopologyManager(
                scenario_name=self.scenario_name,
                n_nodes=self.n_nodes,
                b_symmetric=True,
                undirected_neighbor_num=3,
            )
            topologymanager.generate_topology()
        elif self.scenario.topology == "star" and self.scenario.federation == "CFL":
            # Create a centralized network
            topologymanager = TopologyManager(scenario_name=self.scenario_name, n_nodes=self.n_nodes, b_symmetric=True)
            topologymanager.generate_server_topology()
        else:
            raise ValueError(f"Unknown topology type: {self.scenario.topology}")

        # Assign nodes to topology
        nodes_ip_port = []
        self.config.participants.sort(key=lambda x: x["device_args"]["idx"])
        for i, node in enumerate(self.config.participants):
            nodes_ip_port.append((
                node["network_args"]["ip"],
                node["network_args"]["port"],
                "undefined",
            ))

        topologymanager.add_nodes(nodes_ip_port)
        return topologymanager

    def start_blockchain(self):
        BlockchainDeployer(
            config_dir=f"{self.config_dir}/blockchain",
            input_dir="/nebula/nebula/addons/blockchain",
        )
        try:
            logging.info("Blockchain is being deployed")
            subprocess.check_call([
                "docker",
                "compose",
                "-f",
                f"{self.config_dir}/blockchain/blockchain-docker-compose.yml",
                "up",
                "--remove-orphans",
                "--force-recreate",
                "-d",
                "--build",
            ])
        except subprocess.CalledProcessError as e:
            logging.exception(
                "Docker Compose failed to start Blockchain, please check if Docker Compose is installed (https://docs.docker.com/compose/install/) and Docker Engine is running."
            )
            raise e

    def start_nodes_docker(self):
        import subprocess

        try:
            # First, get the list of IDs of exited containers
            result_ps = subprocess.run(
                "docker ps -aq -f status=exited --filter 'name=nebula'",
                shell=True,
                check=True,
                capture_output=True,
                text=True,
            )

            # Get the container IDs
            container_ids = result_ps.stdout.strip()

            if container_ids:
                # Run the command to remove the containers
                result_rm = subprocess.run(
                    "docker rm $(docker ps -aq -f status=exited --filter 'name=nebula')",
                    shell=True,
                    check=True,
                    capture_output=True,
                    text=True,
                )
                print(f"Dangling containers removed successfully: {result_rm.stdout.strip()}.")
            else:
                print("No dangling containers to remove.")
        except subprocess.CalledProcessError as e:
            print(f"Error removing stopped containers: {e}")
            print(f"Error output: {e.stderr}")
        except Exception as e:
            print(f"Unexpected error: {e}")

        logging.info("Starting nodes using Docker Compose...")
        logging.info(f"env path: {self.env_path}")

        docker_compose_template = textwrap.dedent(
            """
            services:
            {}
        """
        )

        participant_template = textwrap.dedent(
            """
            participant{}:
                image: nebula-core
                restart: no
                volumes:
                    - {}:/nebula
                    - /var/run/docker.sock:/var/run/docker.sock
                extra_hosts:
                    - "host.docker.internal:host-gateway"
                ipc: host
                privileged: true
                command:
                    - /bin/bash
                    - -c
                    - |
                        {} && ifconfig && echo '{} host.docker.internal' >> /etc/hosts && python3.11 /nebula/nebula/node.py {}
                networks:
                    nebula-net-scenario:
                        ipv4_address: {}
                    nebula-net-base:
                    {}
        """
        )
        participant_template = textwrap.indent(participant_template, " " * 4)

        participant_gpu_template = textwrap.dedent(
            """
            participant{}:
                image: nebula-core
                environment:
                    - NVIDIA_DISABLE_REQUIRE=true
                restart: no
                volumes:
                    - {}:/nebula
                    - /var/run/docker.sock:/var/run/docker.sock
                extra_hosts:
                    - "host.docker.internal:host-gateway"
                ipc: host
                privileged: true
                command:
                    - /bin/bash
                    - -c
                    - |
                        {} && ifconfig && echo '{} host.docker.internal' >> /etc/hosts && python3.11 /nebula/nebula/node.py {}
                deploy:
                    resources:
                        reservations:
                            devices:
                                - driver: nvidia
                                  count: all
                                  capabilities: [gpu]
                networks:
                    nebula-net-scenario:
                        ipv4_address: {}
                    nebula-net-base:
                    {}
        """
        )
        participant_gpu_template = textwrap.indent(participant_gpu_template, " " * 4)

        network_template = textwrap.dedent(
            """
            networks:
                nebula-net-scenario:
                    name: nebula-net-scenario
                    driver: bridge
                    ipam:
                        config:
                            - subnet: {}
                              gateway: {}
                nebula-net-base:
                    name: nebula-net-base
                    external: true
                {}
                    {}
                    {}
        """
        )

        # Generate the Docker Compose file dynamically
        services = ""
        self.config.participants.sort(key=lambda x: x["device_args"]["idx"])
        for node in self.config.participants:
            idx = node["device_args"]["idx"]
            path = f"/nebula/app/config/{self.scenario_name}/participant_{idx}.json"
            logging.info(f"Starting node {idx} with configuration {path}")
            logging.info("Node {} is listening on ip {}".format(idx, node["network_args"]["ip"]))
            # Add one service for each participant
            if node["device_args"]["accelerator"] == "gpu":
                logging.info(f"Node {idx} is using GPU")
                services += participant_gpu_template.format(
                    idx,
                    self.root_path,
                    "sleep 10" if node["device_args"]["start"] else "sleep 0",
                    self.scenario.network_gateway,
                    path,
                    node["network_args"]["ip"],
                    "proxy:" if self.scenario.deployment and self.use_blockchain else "",
                )
            else:
                logging.info(f"Node {idx} is using CPU")
                services += participant_template.format(
                    idx,
                    self.root_path,
                    "sleep 10" if node["device_args"]["start"] else "sleep 0",
                    self.scenario.network_gateway,
                    path,
                    node["network_args"]["ip"],
                    "proxy:" if self.scenario.deployment and self.use_blockchain else "",
                )
        docker_compose_file = docker_compose_template.format(services)
        docker_compose_file += network_template.format(
            self.scenario.network_subnet,
            self.scenario.network_gateway,
            "proxy:" if self.scenario.deployment and self.use_blockchain else "",
            "name: chainnet" if self.scenario.deployment and self.use_blockchain else "",
            "external: true" if self.scenario.deployment and self.use_blockchain else "",
        )
        # Write the Docker Compose file in config directory
        with open(f"{self.config_dir}/docker-compose.yml", "w") as f:
            f.write(docker_compose_file)

        # Include additional config to the participants
        for idx, node in enumerate(self.config.participants):
            node["tracking_args"]["log_dir"] = "/nebula/app/logs"
            node["tracking_args"]["config_dir"] = f"/nebula/app/config/{self.scenario_name}"
            node["scenario_args"]["controller"] = self.controller
            node["security_args"]["certfile"] = f"/nebula/app/certs/participant_{node['device_args']['idx']}_cert.pem"
            node["security_args"]["keyfile"] = f"/nebula/app/certs/participant_{node['device_args']['idx']}_key.pem"
            node["security_args"]["cafile"] = "/nebula/app/certs/ca_cert.pem"

            # Write the config file in config directory
            with open(f"{self.config_dir}/participant_{node['device_args']['idx']}.json", "w") as f:
                json.dump(node, f, indent=4)

        # Start the Docker Compose file, catch error if any
        try:
            subprocess.check_call([
                "docker",
                "compose",
                "-f",
                f"{self.config_dir}/docker-compose.yml",
                "up",
                "--build",
                "-d",
            ])
        except subprocess.CalledProcessError:
            raise Exception(
                "Docker Compose failed to start, please check if Docker Compose is installed (https://docs.docker.com/compose/install/) and Docker Engine is running."
            )

        container_ids = None
        logging.info("Waiting for nodes to start...")
        # Loop until all containers are running (equivalent to the number of participants)
        while container_ids is None or len(container_ids) != len(self.config.participants):
            time.sleep(3)
            try:
                # Obtain docker ids
                result = subprocess.run(
                    [
                        "docker",
                        "compose",
                        "-f",
                        f"{self.config_dir}/docker-compose.yml",
                        "ps",
                        "-q",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True,
                )

                if result.returncode != 0:
                    raise Exception(f"Error obtaining docker IDs: {result.stderr}")

                container_ids = result.stdout.strip().split("\n")

            except subprocess.CalledProcessError:
                raise Exception(
                    "Docker Compose failed to start, please check if Docker Compose is installed "
                    "(https://docs.docker.com/compose/install/) and Docker Engine is running."
                )

        # Change log and config directory in dockers to /nebula/app, and change controller endpoint
        for idx, node in enumerate(self.config.participants):
            # Assign docker ID to node
            node["device_args"]["docker_id"] = container_ids[idx]

            # Write the config file in config directory
            with open(f"{self.config_dir}/participant_{node['device_args']['idx']}.json", "w") as f:
                json.dump(node, f, indent=4)

    def start_nodes_process(self):
        logging.info("Starting nodes as processes...")
        logging.info(f"env path: {self.env_path}")

        # Include additional config to the participants
        for idx, node in enumerate(self.config.participants):
            node["tracking_args"]["log_dir"] = os.path.join(self.root_path, "app", "logs")
            node["tracking_args"]["config_dir"] = os.path.join(self.root_path, "app", "config", self.scenario_name)
            node["scenario_args"]["controller"] = self.controller
            node["security_args"]["certfile"] = os.path.join(
                self.root_path,
                "app",
                "certs",
                f"participant_{node['device_args']['idx']}_cert.pem",
            )
            node["security_args"]["keyfile"] = os.path.join(
                self.root_path,
                "app",
                "certs",
                f"participant_{node['device_args']['idx']}_key.pem",
            )
            node["security_args"]["cafile"] = os.path.join(self.root_path, "app", "certs", "ca_cert.pem")

            # Write the config file in config directory
            with open(f"{self.config_dir}/participant_{node['device_args']['idx']}.json", "w") as f:
                json.dump(node, f, indent=4)

        try:
            if self.host_platform == "windows":
                commands = """
                $ParentDir = Split-Path -Parent $PSScriptRoot
                $PID_FILE = "$PSScriptRoot\\current_scenario_pids.txt"
                New-Item -Path $PID_FILE -Force -ItemType File

                """
                sorted_participants = sorted(
                    self.config.participants,
                    key=lambda node: node["device_args"]["idx"],
                    reverse=True,
                )
                for node in sorted_participants:
                    if node["device_args"]["start"]:
                        commands += "Start-Sleep -Seconds 10\n"
                    else:
                        commands += "Start-Sleep -Seconds 2\n"

                    commands += f'Write-Host "Running node {node["device_args"]["idx"]}..."\n'
                    commands += f'$OUT_FILE = "{self.root_path}\\app\\logs\\{self.scenario_name}\\participant_{node["device_args"]["idx"]}.out"\n'
                    commands += f'$ERROR_FILE = "{self.root_path}\\app\\logs\\{self.scenario_name}\\participant_{node["device_args"]["idx"]}.err"\n'

                    # Use Start-Process for executing Python in background and capture PID
                    commands += f"""$process = Start-Process -FilePath "python" -ArgumentList "{self.root_path}\\nebula\\node.py {self.root_path}\\app\\config\\{self.scenario_name}\\participant_{node["device_args"]["idx"]}.json" -PassThru -NoNewWindow -RedirectStandardOutput $OUT_FILE -RedirectStandardError $ERROR_FILE
                Add-Content -Path $PID_FILE -Value $process.Id
                """

                commands += 'Write-Host "All nodes started. PIDs stored in $PID_FILE"\n'

                with open("/nebula/app/config/current_scenario_commands.ps1", "w") as f:
                    f.write(commands)
                os.chmod("/nebula/app/config/current_scenario_commands.ps1", 0o755)
            else:
                commands = '#!/bin/bash\n\nPID_FILE="$(dirname "$0")/current_scenario_pids.txt"\n\n> $PID_FILE\n\n'
                sorted_participants = sorted(
                    self.config.participants,
                    key=lambda node: node["device_args"]["idx"],
                    reverse=True,
                )
                for node in sorted_participants:
                    if node["device_args"]["start"]:
                        commands += "sleep 10\n"
                    else:
                        commands += "sleep 2\n"
                    commands += f'echo "Running node {node["device_args"]["idx"]}..."\n'
                    commands += f"OUT_FILE={self.root_path}/app/logs/{self.scenario_name}/participant_{node['device_args']['idx']}.out\n"
                    commands += f"python3.11 {self.root_path}/nebula/node.py {self.root_path}/app/config/{self.scenario_name}/participant_{node['device_args']['idx']}.json > $OUT_FILE 2>&1 &\n"
                    commands += "echo $! >> $PID_FILE\n\n"

                commands += 'echo "All nodes started. PIDs stored in $PID_FILE"\n'

                with open("/nebula/app/config/current_scenario_commands.sh", "w") as f:
                    f.write(commands)
                os.chmod("/nebula/app/config/current_scenario_commands.sh", 0o755)

        except Exception as e:
            raise Exception(f"Error starting nodes as processes: {e}")

    @classmethod
    def remove_files_by_scenario(cls, scenario_name):
        try:
            shutil.rmtree(Utils.check_path(os.environ["NEBULA_CONFIG_DIR"], scenario_name))
        except FileNotFoundError:
            logging.warning("Files not found, nothing to remove")
        except Exception as e:
            logging.exception("Unknown error while removing files")
            logging.exception(e)
            raise e
        try:
            shutil.rmtree(Utils.check_path(os.environ["NEBULA_LOGS_DIR"], scenario_name))
        except PermissionError:
            # Avoid error if the user does not have enough permissions to remove the tf.events files
            logging.warning("Not enough permissions to remove the files, moving them to tmp folder")
            os.makedirs(
                Utils.check_path(os.environ["NEBULA_ROOT"], os.path.join("app", "tmp", scenario_name)),
                exist_ok=True,
            )
            os.chmod(
                Utils.check_path(os.environ["NEBULA_ROOT"], os.path.join("app", "tmp", scenario_name)),
                0o777,
            )
            shutil.move(
                Utils.check_path(os.environ["NEBULA_LOGS_DIR"], scenario_name),
                Utils.check_path(os.environ["NEBULA_ROOT"], os.path.join("app", "tmp", scenario_name)),
            )
        except FileNotFoundError:
            logging.warning("Files not found, nothing to remove")
        except Exception as e:
            logging.exception("Unknown error while removing files")
            logging.exception(e)
            raise e

    def scenario_finished(self, timeout_seconds):
        client = docker.from_env()
        all_containers = client.containers.list(all=True)
        containers = [container for container in all_containers if self.scenario_name.lower() in container.name.lower()]

        start_time = datetime.now()
        while True:
            all_containers_finished = True
            for container in containers:
                container.reload()
                if container.status != "exited":
                    all_containers_finished = False
                    break
            if all_containers_finished:
                return True

            current_time = datetime.now()
            elapsed_time = current_time - start_time
            if elapsed_time.total_seconds() >= timeout_seconds:
                for container in containers:
                    container.stop()
                return False

            time.sleep(5)



import glob
import json
import logging
import os
import pickle
import shutil
from json import JSONDecodeError

import numpy as np
import pandas as pd

from nebula.addons.trustworthiness.calculation import (
    get_avg_loss_accuracy,
    get_bytes_models,
    get_bytes_sent_recv,
    get_clever_score,
    get_cv,
    get_elapsed_time,
    get_feature_importance_cv,
)
from nebula.addons.trustworthiness.utils import check_field_filled, count_class_samples, get_entropy, read_csv
from nebula.core.models.mnist.cnn import CIFAR10ModelCNN, CIFAR10TorchModelCNN, MNISTModelCNN, MNISTTorchModelCNN
from nebula.core.models.mnist.mlp import MNISTModelMLP, MNISTTorchModelMLP, SyscallModelMLP, SyscallTorchModelMLP

dirname = os.path.dirname(__file__)

logger = logging.getLogger(__name__)


class Factsheet:
    def __init__(self):
        """
        Manager class to populate the FactSheet
        """
        self.factsheet_file_nm = "factsheet.json"
        self.factsheet_template_file_nm = "factsheet_template.json"

    def populate_factsheet_pre_train(self, data, scenario_name):
        """
        Populates the factsheet with values before the training.

        Args:
            data (dict): Contains the data from the scenario.
            scenario_name (string): The name of the scenario.
        """

        factsheet_file = os.path.join(dirname, f"files/{scenario_name}/{self.factsheet_file_nm}")

        factsheet_template = os.path.join(dirname, f"configs/{self.factsheet_template_file_nm}")

        if not os.path.exists(factsheet_file):
            shutil.copyfile(factsheet_template, factsheet_file)

        with open(factsheet_file, "r+") as f:
            factsheet = {}

            try:
                factsheet = json.load(f)

                if data is not None:
                    logger.info("FactSheet: Populating factsheet with pre training metrics")

                    federation = data["federation"]
                    n_nodes = int(data["n_nodes"])
                    dataset = data["dataset"]
                    algorithm = data["model"]
                    aggregation_algorithm = data["agg_algorithm"]
                    n_rounds = int(data["rounds"])
                    attack = data["attacks"]
                    poisoned_node_percent = int(data["poisoned_node_percent"])
                    poisoned_sample_percent = int(data["poisoned_sample_percent"])
                    poisoned_noise_percent = int(data["poisoned_noise_percent"])
                    with_reputation = data["with_reputation"]
                    is_dynamic_topology = data["is_dynamic_topology"]
                    is_dynamic_aggregation = data["is_dynamic_aggregation"]
                    target_aggregation = data["target_aggregation"]

                    if attack != "No Attack" and with_reputation == True and is_dynamic_aggregation == True:
                        background = f"For the project setup, the most important aspects are the following: The federation architecture is {federation}, involving {n_nodes} clients, the dataset used is {dataset}, the learning algorithm is {algorithm}, the aggregation algorithm is {aggregation_algorithm} and the number of rounds is {n_rounds}. In addition, the type of attack used against the clients is {attack}, where the percentage of attacked nodes is {poisoned_node_percent}, the percentage of attacked samples of each node is {poisoned_sample_percent}, and the percent of poisoned noise is {poisoned_noise_percent}. A reputation-based defence with a dynamic aggregation based on the aggregation algorithm {target_aggregation} is used, and the trustworthiness of the project is desired."

                    elif attack != "No Attack" and with_reputation == True and is_dynamic_topology == True:
                        background = f"For the project setup, the most important aspects are the following: The federation architecture is {federation}, involving {n_nodes} clients, the dataset used is {dataset}, the learning algorithm is {algorithm}, the aggregation algorithm is {aggregation_algorithm} and the number of rounds is {n_rounds}. In addition, the type of attack used against the clients is {attack}, where the percentage of attacked nodes is {poisoned_node_percent}, the percentage of attacked samples of each node is {poisoned_sample_percent}, and the percent of poisoned noise is {poisoned_noise_percent}. A reputation-based defence with a dynamic topology is used, and the trustworthiness of the project is desired."

                    elif attack != "No Attack" and with_reputation == False:
                        background = f"For the project setup, the most important aspects are the following: The federation architecture is {federation}, involving {n_nodes} clients, the dataset used is {dataset}, the learning algorithm is {algorithm}, the aggregation algorithm is {aggregation_algorithm} and the number of rounds is {n_rounds}. In addition, the type of attack used against the clients is {attack}, where the percentage of attacked nodes is {poisoned_node_percent}, the percentage of attacked samples of each node is {poisoned_sample_percent}, and the percent of poisoned noise is {poisoned_noise_percent}. No defence mechanism is used, and the trustworthiness of the project is desired."

                    elif attack == "No Attack":
                        background = f"For the project setup, the most important aspects are the following: The federation architecture is {federation}, involving {n_nodes} clients, the dataset used is {dataset}, the learning algorithm is {algorithm}, the aggregation algorithm is {aggregation_algorithm} and the number of rounds is {n_rounds}. No attacks against clients are used, and the trustworthiness of the project is desired."

                    # Set project specifications
                    factsheet["project"]["overview"] = data["scenario_title"]
                    factsheet["project"]["purpose"] = data["scenario_description"]
                    factsheet["project"]["background"] = background

                    # Set data specifications
                    factsheet["data"]["provenance"] = data["dataset"]
                    factsheet["data"]["preprocessing"] = data["topology"]

                    # Set participants
                    factsheet["participants"]["client_num"] = data["n_nodes"] or ""
                    factsheet["participants"]["sample_client_rate"] = 1
                    factsheet["participants"]["client_selector"] = ""

                    # Set configuration
                    factsheet["configuration"]["aggregation_algorithm"] = data["agg_algorithm"] or ""
                    factsheet["configuration"]["training_model"] = data["model"] or ""
                    factsheet["configuration"]["personalization"] = False
                    factsheet["configuration"]["visualization"] = True
                    factsheet["configuration"]["total_round_num"] = n_rounds

                    if poisoned_noise_percent != 0:
                        factsheet["configuration"]["differential_privacy"] = True
                        factsheet["configuration"]["dp_epsilon"] = poisoned_noise_percent
                    else:
                        factsheet["configuration"]["differential_privacy"] = False
                        factsheet["configuration"]["dp_epsilon"] = ""

                    if dataset == "MNIST" and algorithm == "MLP":
                        model = MNISTModelMLP()
                    elif dataset == "MNIST" and algorithm == "CNN":
                        model = MNISTModelCNN()
                    elif dataset == "Syscall" and algorithm == "MLP":
                        model = SyscallModelMLP()
                    else:
                        model = CIFAR10ModelCNN()

                    factsheet["configuration"]["learning_rate"] = model.get_learning_rate()
                    factsheet["configuration"]["trainable_param_num"] = model.count_parameters()
                    factsheet["configuration"]["local_update_steps"] = 1

            except JSONDecodeError as e:
                logger.warning(f"{factsheet_file} is invalid")
                logger.error(e)

            f.seek(0)
            f.truncate()
            json.dump(factsheet, f, indent=4)
            f.close()

    def populate_factsheet_post_train(self, scenario):
        """
        Populates the factsheet with values after the training.

        Args:
            scenario (object): The scenario object.
        """
        scenario_name = scenario[0]

        factsheet_file = os.path.join(dirname, f"files/{scenario_name}/{self.factsheet_file_nm}")

        logger.info("FactSheet: Populating factsheet with post training metrics")

        with open(factsheet_file, "r+") as f:
            factsheet = {}
            try:
                factsheet = json.load(f)

                dataset = factsheet["data"]["provenance"]
                model = factsheet["configuration"]["training_model"]

                actual_dir = os.getcwd()
                files_dir = f"{actual_dir}/trustworthiness/files/{scenario_name}"
                data_dir = f"{actual_dir}/trustworthiness/data/"

                models_files = glob.glob(os.path.join(files_dir, "*final_model*"))
                bytes_sent_files = glob.glob(os.path.join(files_dir, "*bytes_sent*"))
                bytes_recv_files = glob.glob(os.path.join(files_dir, "*bytes_recv*"))
                loss_files = glob.glob(os.path.join(files_dir, "*loss*"))
                accuracy_files = glob.glob(os.path.join(files_dir, "*accuracy*"))
                dataloaders_files = glob.glob(os.path.join(files_dir, "*train_loader*"))
                test_dataloader_file = f"{files_dir}/participant_1_test_loader.pk"
                train_model_file = f"{files_dir}/participant_1_train_model.pk"
                emissions_file = os.path.join(files_dir, "emissions.csv")

                # Entropy
                i = 0
                for file in dataloaders_files:
                    with open(file, "rb") as file:
                        dataloader = pickle.load(file)
                    get_entropy(i, scenario_name, dataloader)
                    i += 1

                with open(f"{files_dir}/entropy.json") as file:
                    entropy_distribution = json.load(file)

                values = np.array(list(entropy_distribution.values()))

                normalized_values = (values - np.min(values)) / (np.max(values) - np.min(values))

                avg_entropy = np.mean(normalized_values)

                factsheet["data"]["avg_entropy"] = avg_entropy

                # Set performance data
                result_avg_loss_accuracy = get_avg_loss_accuracy(loss_files, accuracy_files)
                factsheet["performance"]["test_loss_avg"] = result_avg_loss_accuracy[0]
                factsheet["performance"]["test_acc_avg"] = result_avg_loss_accuracy[1]
                test_acc_cv = get_cv(std=result_avg_loss_accuracy[2], mean=result_avg_loss_accuracy[1])
                factsheet["fairness"]["test_acc_cv"] = 1 if test_acc_cv > 1 else test_acc_cv

                factsheet["system"]["avg_time_minutes"] = get_elapsed_time(scenario)
                factsheet["system"]["avg_model_size"] = get_bytes_models(models_files)

                result_bytes_sent_recv = get_bytes_sent_recv(bytes_sent_files, bytes_recv_files)
                factsheet["system"]["total_upload_bytes"] = result_bytes_sent_recv[0]
                factsheet["system"]["total_download_bytes"] = result_bytes_sent_recv[1]
                factsheet["system"]["avg_upload_bytes"] = result_bytes_sent_recv[2]
                factsheet["system"]["avg_download_bytes"] = result_bytes_sent_recv[3]

                factsheet["fairness"]["selection_cv"] = 1

                count_class_samples(scenario_name, dataloaders_files)

                with open(f"{files_dir}/count_class.json") as file:
                    class_distribution = json.load(file)

                class_samples_sizes = [x for x in class_distribution.values()]
                class_imbalance = get_cv(list=class_samples_sizes)
                factsheet["fairness"]["class_imbalance"] = 1 if class_imbalance > 1 else class_imbalance

                with open(train_model_file, "rb") as file:
                    lightning_model = pickle.load(file)

                if dataset == "MNIST" and model == "MLP":
                    pytorch_model = MNISTTorchModelMLP()
                elif dataset == "MNIST" and model == "CNN":
                    pytorch_model = MNISTTorchModelCNN()
                elif dataset == "Syscall" and model == "MLP":
                    pytorch_model = SyscallTorchModelMLP()
                else:
                    pytorch_model = CIFAR10TorchModelCNN()

                pytorch_model.load_state_dict(lightning_model.state_dict())

                with open(test_dataloader_file, "rb") as file:
                    test_dataloader = pickle.load(file)

                test_sample = next(iter(test_dataloader))

                lr = factsheet["configuration"]["learning_rate"]
                value_clever = get_clever_score(pytorch_model, test_sample, 10, lr)

                factsheet["performance"]["test_clever"] = 1 if value_clever > 1 else value_clever

                feature_importance = get_feature_importance_cv(pytorch_model, test_sample)

                factsheet["performance"]["test_feature_importance_cv"] = (
                    1 if feature_importance > 1 else feature_importance
                )

                # Set emissions metrics
                emissions = None if emissions_file is None else read_csv(emissions_file)
                if emissions is not None:
                    logger.info("FactSheet: Populating emissions")
                    cpu_spez_df = pd.read_csv(os.path.join(data_dir, "CPU_benchmarks_v4.csv"), header=0)
                    emissions["CPU_model"] = (
                        emissions["CPU_model"].astype(str).str.replace(r"\([^)]*\)", "", regex=True)
                    )
                    emissions["CPU_model"] = emissions["CPU_model"].astype(str).str.replace(r" CPU", "", regex=True)
                    emissions["GPU_model"] = emissions["GPU_model"].astype(str).str.replace(r"[0-9] x ", "", regex=True)
                    emissions = pd.merge(
                        emissions,
                        cpu_spez_df[["cpuName", "powerPerf"]],
                        left_on="CPU_model",
                        right_on="cpuName",
                        how="left",
                    )
                    gpu_spez_df = pd.read_csv(os.path.join(data_dir, "GPU_benchmarks_v7.csv"), header=0)
                    emissions = pd.merge(
                        emissions,
                        gpu_spez_df[["gpuName", "powerPerformance"]],
                        left_on="GPU_model",
                        right_on="gpuName",
                        how="left",
                    )

                    emissions.drop("cpuName", axis=1, inplace=True)
                    emissions.drop("gpuName", axis=1, inplace=True)
                    emissions["powerPerf"] = emissions["powerPerf"].astype(float)
                    emissions["powerPerformance"] = emissions["powerPerformance"].astype(float)
                    client_emissions = emissions.loc[emissions["role"] == "client"]
                    client_avg_carbon_intensity = round(client_emissions["energy_grid"].mean(), 2)
                    factsheet["sustainability"]["avg_carbon_intensity_clients"] = check_field_filled(
                        factsheet,
                        ["sustainability", "avg_carbon_intensity_clients"],
                        client_avg_carbon_intensity,
                        "",
                    )
                    factsheet["sustainability"]["emissions_training"] = check_field_filled(
                        factsheet,
                        ["sustainability", "emissions_training"],
                        client_emissions["emissions"].sum(),
                        "",
                    )
                    factsheet["participants"]["avg_dataset_size"] = check_field_filled(
                        factsheet,
                        ["participants", "avg_dataset_size"],
                        client_emissions["sample_size"].mean(),
                        "",
                    )

                    server_emissions = emissions.loc[emissions["role"] == "server"]
                    server_avg_carbon_intensity = round(server_emissions["energy_grid"].mean(), 2)
                    factsheet["sustainability"]["avg_carbon_intensity_server"] = check_field_filled(
                        factsheet,
                        ["sustainability", "avg_carbon_intensity_server"],
                        server_avg_carbon_intensity,
                        "",
                    )
                    factsheet["sustainability"]["emissions_aggregation"] = check_field_filled(
                        factsheet,
                        ["sustainability", "emissions_aggregation"],
                        server_emissions["emissions"].sum(),
                        "",
                    )
                    GPU_powerperf = (server_emissions.loc[server_emissions["GPU_used"] == True])["powerPerformance"]
                    CPU_powerperf = (server_emissions.loc[server_emissions["CPU_used"] == True])["powerPerf"]
                    server_power_performance = round(pd.concat([GPU_powerperf, CPU_powerperf]).mean(), 2)
                    factsheet["sustainability"]["avg_power_performance_server"] = check_field_filled(
                        factsheet,
                        ["sustainability", "avg_power_performance_server"],
                        server_power_performance,
                        "",
                    )

                    GPU_powerperf = (client_emissions.loc[client_emissions["GPU_used"] == True])["powerPerformance"]
                    CPU_powerperf = (client_emissions.loc[client_emissions["CPU_used"] == True])["powerPerf"]
                    clients_power_performance = round(pd.concat([GPU_powerperf, CPU_powerperf]).mean(), 2)
                    factsheet["sustainability"]["avg_power_performance_clients"] = clients_power_performance

                    factsheet["sustainability"]["emissions_communication_uplink"] = check_field_filled(
                        factsheet,
                        ["sustainability", "emissions_communication_uplink"],
                        factsheet["system"]["total_upload_bytes"]
                        * 2.24e-10
                        * factsheet["sustainability"]["avg_carbon_intensity_clients"],
                        "",
                    )
                    factsheet["sustainability"]["emissions_communication_downlink"] = check_field_filled(
                        factsheet,
                        ["sustainability", "emissions_communication_downlink"],
                        factsheet["system"]["total_download_bytes"]
                        * 2.24e-10
                        * factsheet["sustainability"]["avg_carbon_intensity_server"],
                        "",
                    )

            except JSONDecodeError as e:
                logger.warning(f"{factsheet_file} is invalid")
                logger.error(e)

            f.seek(0)
            f.truncate()
            json.dump(factsheet, f, indent=4)
            f.close()
